import logging
from copy import deepcopy

from numpy import argsort as numpy_argsort
from numpy import atleast_1d as numpy_atleast_1d
from numpy import ndarray as numpy_ndarray

from .decorators import (
    _deprecated_kwarg_check,
    _display_or_return,
    _manage_log_level_via_verbosity,
)
from .functions import atol as cf_atol
from .functions import equals as cf_equals
from .functions import inspect as cf_inspect
from .functions import rtol as cf_rtol

logger = logging.getLogger(__name__)


class Flags:
    """Self-describing CF flag values.

    Stores the flag_values, flag_meanings and flag_masks CF attributes
    in an internally consistent manner.

    """

    def __init__(self, **kwargs):
        """**Initialization**

        :Parameters:

            flag_values : optional
                The flag_values CF property. Sets the `flag_values`
                attribute.

            flag_meanings : optional
                The flag_meanings CF property. Sets the `flag_meanings`
                attribute.

            flag_masks : optional
                The flag_masks CF property. Sets the `flag_masks`
                attribute.

        """
        for attr, value in kwargs.items():
            if value is not None:
                setattr(self, attr, value)

    def __eq__(self, other):
        """x.__eq__(y) <==> x==y <==> x.equals(y)"""
        return self.equals(other)

    def __ne__(self, other):
        """x.__ne__(y) <==> x!=y <==> not x.equals(y)"""
        return not self.equals(other)

    def __hash__(self):
        """Return the hash value of the flags.

        Note that the flags will be sorted in place.

        :Returns:

            `int`
                The hash value.

        **Examples**

        >>> hash(f)
        -956218661958673979

        """
        self.sort()

        x = [
            tuple(getattr(self, attr, ()))
            for attr in ("_flag_meanings", "_flag_values", "_flag_masks")
        ]

        return hash(tuple(x))

    def __bool__(self):
        """x.__bool__() <==> x!=0."""
        for attr in ("_flag_meanings", "_flag_values", "_flag_masks"):
            if hasattr(self, attr):
                return True

        return False

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def flag_values(self):
        """The flag_values CF attribute.

        Stored as a 1-d numpy array but may be set as any array-like
        object.

        **Examples**

        >>> f.flag_values = ['a', 'b', 'c']
        >>> f.flag_values
        array(['a', 'b', 'c'], dtype='|S1')
        >>> f.flag_values = numpy.arange(4, dtype='int8')
        >>> f.flag_values
        array([1, 2, 3, 4], dtype=int8)
        >>> f.flag_values = 1
        >>> f.flag_values
        array([1])

        """
        try:
            return self._flag_values
        except AttributeError:
            raise AttributeError(
                "'%s' has no attribute 'flag_values'" % self.__class__.__name__
            )

    @flag_values.setter
    def flag_values(self, value):
        if not isinstance(value, numpy_ndarray):
            value = numpy_atleast_1d(value)
        self._flag_values = value

    @flag_values.deleter
    def flag_values(self):
        try:
            del self._flag_values
        except AttributeError:
            raise AttributeError(
                "Can't delete '%s' attribute 'flag_values'"
                % self.__class__.__name__
            )

    # ----------------------------------------------------------------
    # Property attribute: flag_masks
    # ----------------------------------------------------------------
    @property
    def flag_masks(self):
        """The flag_masks CF attribute.

        Stored as a 1-d numpy array but may be set as array-like object.

        **Examples**

        >>> f.flag_masks = numpy.array([1, 2, 4], dtype='int8')
        >>> f.flag_masks
        array([1, 2, 4], dtype=int8)
        >>> f.flag_masks = 1
        >>> f.flag_masks
        array([1])

        """
        try:
            return self._flag_masks
        except AttributeError:
            raise AttributeError(
                "'%s' object has no attribute 'flag_masks'"
                % self.__class__.__name__
            )

    @flag_masks.setter
    def flag_masks(self, value):
        if not isinstance(value, numpy_ndarray):
            value = numpy_atleast_1d(value)

        self._flag_masks = value

    @flag_masks.deleter
    def flag_masks(self):
        try:
            del self._flag_masks
        except AttributeError:
            raise AttributeError(
                "Can't delete '%s' attribute 'flag_masks'"
                % self.__class__.__name__
            )

    @property
    def flag_meanings(self):
        """The flag_meanings CF attribute.

        Stored as a 1-d numpy string array but may be set as a space
        delimited string or any array-like object.

        **Examples**

        >>> f.flag_meanings = 'low medium      high'
        >>> f.flag_meanings
        array(['low', 'medium', 'high'],
              dtype='|S6')
        >>> f.flag_meanings = ['left', 'right']
        >>> f.flag_meanings
        array(['left', 'right'],
              dtype='|S5')
        >>> f.flag_meanings = 'ok'
        >>> f.flag_meanings
        array(['ok'],
              dtype='|S2')
        >>> f.flag_meanings = numpy.array(['a', 'b'])
        >>> f.flag_meanings
        array(['a', 'b'],
              dtype='|S1')

        """
        try:
            return self._flag_meanings
        except AttributeError:
            raise AttributeError(
                "'%s' object has no attribute 'flag_meanings'"
                % self.__class__.__name__
            )

    @flag_meanings.setter
    def flag_meanings(self, value):
        if isinstance(value, str):
            value = numpy_atleast_1d(value.split())
        elif not isinstance(value, numpy_ndarray):
            value = numpy_atleast_1d(value)

        self._flag_meanings = value

    @flag_meanings.deleter
    def flag_meanings(self):
        try:
            del self._flag_meanings
        except AttributeError:
            raise AttributeError(
                "Can't delete '%s' attribute 'flag_meanings'"
                % self.__class__.__name__
            )

    def __repr__(self):
        """x.__repr__() <==> repr(x)"""
        string = []
        if hasattr(self, "flag_values"):
            string.append("flag_values=%s" % str(self.flag_values))

        if hasattr(self, "flag_masks"):
            string.append("flag_masks=%s" % str(self.flag_masks))

        if hasattr(self, "flag_meanings"):
            string.append("flag_meanings=%s" % str(self.flag_meanings))

        return "<CF %s: %s>" % (self.__class__.__name__, ", ".join(string))

    def copy(self):
        """Return a deep copy.

        Equivalent to ``copy.deepcopy(f)``

        :Returns:

                The deep copy.

        **Examples**

        >>> f.copy()

        """
        return deepcopy(self)

    @_display_or_return
    def dump(self, display=True, _level=0):
        """Return a string containing a full description of the
        instance.

        :Parameters:

            display : bool, optional
                If False then return the description as a string. By
                default the description is printed, i.e. ``f.dump()`` is
                equivalent to ``print(f.dump(display=False))``.

        :Returns:

            `None` or `str`
                A string containing the description.

        """
        indent0 = "    " * _level
        indent1 = "    " * (_level + 1)

        string = ["%sFlags:" % indent0]

        for attr in ("_flag_values", "_flag_meanings", "_flag_masks"):
            value = getattr(self, attr, None)
            if value is not None:
                string.append("%s%s = %s" % (indent1, attr[1:], list(value)))

        return "\n".join(string)

    @_deprecated_kwarg_check("traceback")
    @_manage_log_level_via_verbosity
    def equals(
        self,
        other,
        rtol=None,
        atol=None,
        ignore_fill_value=False,
        verbose=None,
        traceback=False,
    ):
        """True if two groups of flags are logically equal, False
        otherwise.

        Note that both instances are sorted in place prior to the comparison.

        :Parameters:

            other:
                The object to compare for equality.

            atol: float, optional
                The absolute tolerance for all numerical comparisons, By
                default the value returned by the `atol` function is used.

            rtol: float, optional
                The relative tolerance for all numerical comparisons, By
                default the value returned by the `rtol` function is used.

            ignore_fill_value: bool, optional
                If True then data arrays with different fill values are
                considered equal. By default they are considered unequal.

            traceback: deprecated at version 3.0.0.
                Use *verbose* instead.

        :Returns:

            `bool`
                Whether or not the two instances are equal.

        **Examples**

        >>> f
        <CF Flags: flag_values=[1 0 2], flag_masks=[2 0 2], flag_meanings=['medium' 'low' 'high']>
        >>> g
        <CF Flags: flag_values=[2 0 1], flag_masks=[2 0 2], flag_meanings=['high' 'low' 'medium']>
        >>> f.equals(g)
        True
        >>> f
        <CF Flags: flag_values=[0 1 2], flag_masks=[0 2 2], flag_meanings=['low' 'medium' 'high']>
        >>> g
        <CF Flags: flag_values=[0 1 2], flag_masks=[0 2 2], flag_meanings=['low' 'medium' 'high']>

        """
        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            logger.info(
                "%s: Different type: %s, %s"
                % (
                    self.__class__.__name__,
                    self.__class__.__name__,
                    other.__class__.__name__,
                )
            )  # pragma: no cover
            return False

        self.sort()
        other.sort()

        # Set default tolerances
        if rtol is None:
            rtol = float(cf_rtol())

        if atol is None:
            atol = float(cf_atol())

        for attr in ("_flag_meanings", "_flag_values", "_flag_masks"):
            if hasattr(self, attr):
                if not hasattr(other, attr):
                    logger.info(
                        "%s: Different attributes: %s"
                        % (self.__class__.__name__, attr[1:])
                    )  # pragma: no cover
                    return False

                x = getattr(self, attr)
                y = getattr(other, attr)

                if x.shape != y.shape or not cf_equals(
                    x,
                    y,
                    rtol=rtol,
                    atol=atol,
                    ignore_fill_value=ignore_fill_value,
                    verbose=verbose,
                ):
                    print(
                        "%s: Different '%s': %r, %r"
                        % (self.__class__.__name__, attr[1:], x, y)
                    )  # pragma: no cover
                    return False

            elif hasattr(other, attr):
                print(
                    "%s: Different attributes: %s"
                    % (self.__class__.__name__, attr[1:])
                )  # pragma: no cover
                return False

        return True

    def inspect(self):
        """Inspect the object for debugging.

        .. seealso:: `cf.inspect`

        :Returns:

            `None`

        """
        print(cf_inspect(self))  # pragma: no cover

    def sort(self):
        """Sort the flags in place.

        By default sort by flag values. If flag values are not present
        then sort by flag meanings. If flag meanings are not present then
        sort by flag_masks.

        :Returns:

            `None`

        **Examples**

        >>> f
        <CF Flags: flag_values=[2 0 1], flag_masks=[2 0 2], flag_meanings=['high' 'low' 'medium']>
        >>> f.sort()
        >>> f
        <CF Flags: flag_values=[0 1 2], flag_masks=[0 2 2], flag_meanings=['low' 'medium' 'high']>

        """
        if not self:
            return

        # Sort all three attributes
        for attr in ("flag_values", "_flag_meanings", "_flag_masks"):
            if hasattr(self, attr):
                indices = numpy_argsort(getattr(self, attr))
                break

        for attr in ("_flag_values", "_flag_meanings", "_flag_masks"):
            if hasattr(self, attr):
                array = getattr(self, attr).view()
                array[...] = array[indices]



import json
from ast import literal_eval as ast_literal_eval

import cfdm
from numpy import dtype as numpy_dtype

from ...constants import _file_to_fh
from ...functions import dirname, pathjoin
from ...units import Units


class NetCDFRead(cfdm.read_write.netcdf.NetCDFRead):
    """A container for instantiating Fields from a netCDF dataset.

    .. versionadded:: 3.0.0

    """

    def _ncdimensions(self, ncvar, ncdimensions=None):
        """Return a list of the netCDF dimensions corresponding to a
        netCDF variable.

        If the variable has been compressed then the *implied
        uncompressed* dimensions are returned.

        For a CFA variable, the netCDF dimensions are taken from the
        'cfa_dimensions' netCDF attribute.

        .. versionadded:: 3.0.0

        :Parameters:

            ncvar: `str`
                The netCDF variable name.

            ncdimensions: sequence of `str`, optional
                Use these netCDF dimensions, rather than retrieving them
                from the netCDF variable itself. This allows the
                dimensions of a domain variable to be parsed. Note that
                this only parameter only needs to be used once because the
                parsed domain dimensions are automatically stored in
                `self.read_var['domain_ncdimensions'][ncvar]`.

                .. versionadded:: 3.11.0

        :Returns:

            `list`
                The netCDF dimension names spanned by the netCDF variable.

        **Examples**

        >>> n._ncdimensions('humidity')
        ['time', 'lat', 'lon']

        """
        g = self.read_vars

        cfa = (
            g["cfa"]
            and ncvar not in g["external_variables"]
            and g["variable_attributes"][ncvar].get("cf_role")
            == "cfa_variable"
        )

        if not cfa:
            return super()._ncdimensions(ncvar, ncdimensions=ncdimensions)

        # Still here? Then we have a CFA variable.
        ncdimensions = (
            g["variable_attributes"][ncvar].get("cfa_dimensions", "").split()
        )

        return list(map(str, ncdimensions))

    def _get_domain_axes(self, ncvar, allow_external=False):
        """Return the domain axis identifiers that correspond to a
        netCDF variable's netCDF dimensions.

        For a CFA variable, the netCDF dimensions are taken from the
        'cfa_dimensions' netCDF attribute.

        :Parameter:

            ncvar: `str`
                The netCDF variable name.

            allow_external: `bool`
                If `True` and *ncvar* is an external variable then return an
                empty list.

        :Returns:

            `list`

        **Examples**

        >>> r._get_domain_axes('areacello')
        ['domainaxis0', 'domainaxis1']

        >>> r._get_domain_axes('areacello', allow_external=True)
        []

        """
        g = self.read_vars

        cfa = (
            g["cfa"]
            and ncvar not in g["external_variables"]
            and g["variable_attributes"][ncvar].get("cf_role")
            == "cfa_variable"
        )

        if not cfa:
            return super()._get_domain_axes(
                ncvar=ncvar, allow_external=allow_external
            )

        # Still here?
        cfa_dimensions = (
            g["variable_attributes"][ncvar].get("cfa_dimensions", "").split()
        )

        ncdim_to_axis = g["ncdim_to_axis"]
        axes = [
            ncdim_to_axis[ncdim]
            for ncdim in cfa_dimensions
            if ncdim in ncdim_to_axis
        ]

        return axes

    def _create_data(
        self,
        ncvar,
        construct=None,
        unpacked_dtype=False,
        uncompress_override=None,
        parent_ncvar=None,
    ):
        """Create data for a netCDF or CFA-netCDF variable.

        .. versionadded:: 3.0.0

        :Parameters:

            ncvar: `str`
                The name of the netCDF variable that contains the data.

            construct: optional

            unpacked_dtype: `False` or `numpy.dtype`, optional

            uncompress_override: `bool`, optional

        :Returns:

            `Data`

        """
        g = self.read_vars

        is_cfa_variable = (
            g["cfa"]
            and construct.get_property("cf_role", None) == "cfa_variable"
        )

        if not is_cfa_variable:
            # --------------------------------------------------------
            # Create data for a normal netCDF variable
            # --------------------------------------------------------
            return super()._create_data(
                ncvar=ncvar,
                construct=construct,
                unpacked_dtype=unpacked_dtype,
                uncompress_override=uncompress_override,
                parent_ncvar=parent_ncvar,
            )

        # ------------------------------------------------------------
        # Still here? Then create data for a CFA netCDF variable
        # ------------------------------------------------------------
        #        print ('    Creating data from CFA variable', repr(ncvar),
        #               repr(construct))
        try:
            cfa_data = json.loads(construct.get_property("cfa_array"))
        except ValueError as error:
            raise ValueError(
                "Error during JSON-decoding of netCDF attribute 'cfa_array': "
                f"{error}"
            )

        variable = g["variables"][ncvar]

        cfa_data["file"] = g["filename"]
        cfa_data["Units"] = construct.Units
        cfa_data["fill_value"] = construct.fill_value()
        cfa_data["_pmshape"] = cfa_data.pop("pmshape", ())
        cfa_data["_pmaxes"] = cfa_data.pop("pmdimensions", ())

        base = cfa_data.get("base", None)
        if base is not None:
            cfa_data["base"] = pathjoin(dirname(g["filename"]), base)

        ncdimensions = construct.get_property("cfa_dimensions", "").split()
        dtype = variable.dtype

        if dtype is str:
            # netCDF string types have a dtype of `str`, which needs
            # to be reset as a numpy.dtype, but we don't know what
            # without reading the data, so set it to None for now.
            dtype = None

        # UNICODE???? TODO
        if self._is_char(ncvar) and dtype.kind in "SU" and ncdimensions:
            strlen = g["nc"].dimensions[ncdimensions[-1]].size
            if strlen > 1:
                ncdimensions.pop()
                dtype = numpy_dtype(f"S{strlen}")

        cfa_data["dtype"] = dtype
        cfa_data["_axes"] = ncdimensions
        cfa_data["shape"] = [
            g["nc"].dimensions[ncdim].size for ncdim in ncdimensions
        ]

        for attrs in cfa_data["Partitions"]:
            # FORMAT
            sformat = attrs.get("subarray", {}).pop("format", "netCDF")
            if sformat is not None:
                attrs["format"] = sformat

            # DTYPE
            dtype = attrs.get("subarray", {}).pop("dtype", None)
            if dtype not in (None, "char"):
                attrs["subarray"]["dtype"] = numpy_dtype(dtype)

            # UNITS and CALENDAR
            units = attrs.pop("punits", None)
            calendar = attrs.pop("pcalendar", None)
            if units is not None or calendar is not None:
                attrs["Units"] = Units(units, calendar)

            # AXES
            pdimensions = attrs.pop("pdimensions", None)
            if pdimensions is not None:
                attrs["axes"] = pdimensions

            # REVERSE
            reverse = attrs.pop("reverse", None)
            if reverse is not None:
                attrs["reverse"] = reverse

            # LOCATION: Change to python indexing (i.e. range does not
            #           include the final index)
            for r in attrs["location"]:
                r[1] += 1

            # PART: Change to python indexing (i.e. slice range does
            #       not include the final index)
            part = attrs.get("part", None)
            if part:
                p = []
                for x in ast_literal_eval(part):
                    if isinstance(x, list):
                        if x[2] > 0:
                            p.append(slice(x[0], x[1] + 1, x[2]))
                        elif x[1] == 0:
                            p.append(slice(x[0], None, x[2]))
                        else:
                            p.append(slice(x[0], x[1] - 1, x[2]))
                    else:
                        p.append(list(x))

                attrs["part"] = p

        construct.del_property("cf_role")
        construct.del_property("cfa_array")
        construct.del_property("cfa_dimensions", None)

        out = self._create_Data(loadd=cfa_data)

        return out

    def _create_Data(
        self,
        array=None,
        units=None,
        calendar=None,
        ncvar=None,
        loadd=None,
        **kwargs,
    ):
        """Create a Data object.

        .. versionadded:: 3.0.0

        :Parameters:

            ncvar: `str`
                The netCDF variable from which to get units and calendar.

        """
        try:
            compressed = array.get_compression_type()  # TODO
        except AttributeError:
            compressed = False

        if not compressed:
            # Do not chunk compressed data (for now ...)
            chunk = False
        else:
            chunk = self.read_vars.get("chunk", True)

        return super()._create_Data(
            array=array,
            units=units,
            calendar=calendar,
            ncvar=ncvar,
            loadd=loadd,
            chunk=chunk,
            **kwargs,
        )

    def _customize_read_vars(self):
        """Customize the read parameters.

        .. versionadded:: 3.0.0

        """
        super()._customize_read_vars()

        g = self.read_vars

        # ------------------------------------------------------------
        # Find out if this is a CFA file
        # ------------------------------------------------------------
        g["cfa"] = "CFA" in g["global_attributes"].get(
            "Conventions", ()
        )  # TODO

        # ------------------------------------------------------------
        # Do not create fields from CFA private variables
        # ------------------------------------------------------------
        if g["cfa"]:
            for ncvar in g["variables"]:
                if (
                    g["variable_attributes"][ncvar].get("cf_role", None)
                    == "cfa_private"
                ):
                    g["do_not_create_field"].add(ncvar)

        # ------------------------------------------------------------
        #
        # ------------------------------------------------------------
        if g["cfa"]:
            for ncvar, ncdims in tuple(g["variable_dimensions"].items()):
                if ncdims != ():
                    continue

                if not (
                    ncvar not in g["external_variables"]
                    and g["variable_attributes"][ncvar].get("cf_role")
                    == "cfa_variable"
                ):
                    continue

                ncdimensions = (
                    g["variable_attributes"][ncvar]
                    .get("cfa_dimensions", "")
                    .split()
                )
                if ncdimensions:
                    g["variable_dimensions"][ncvar] = tuple(
                        map(str, ncdimensions)
                    )

    def file_open(self, filename, flatten=True, verbose=None):
        """Open the netCDf file for reading.

        :Paramters:

            filename: `str`
                The netCDF file to be read.

            flatten: `bool`, optional
                If False then do not flatten a grouped file. Ignored if
                the file has no groups.

                .. versionadded:: 3.6.0

        :Returns:

            `netCDF4.Dataset`
                The object for the file.

        """
        out = super().file_open(filename, flatten=flatten, verbose=verbose)
        _file_to_fh["netCDF"].pop(filename, None)
        return out



import logging

import cfdm

from .query import Query

logger = logging.getLogger(__name__)


class Constructs(cfdm.Constructs):
    """A container for metadata constructs.

    The following metadata constructs can be included:

    * auxiliary coordinate constructs
    * coordinate reference constructs
    * cell measure constructs
    * dimension coordinate constructs
    * domain ancillary constructs
    * domain axis constructs
    * cell method constructs
    * field ancillary constructs

    The container may be used by `Field` and `Domain` instances. In
    the latter case cell method and field ancillary constructs must be
    flagged as "ignored" (see the *_ignore* parameter).

    The container is like a dictionary in many ways, in that it stores
    key/value pairs where the key is the unique construct key with
    corresponding metadata construct value, and provides some of the
    usual dictionary methods.

    **Calling**

    Calling a `Constructs` instance selects metadata constructs by
    identity and is an alias for the `filter_by_identity` method. For
    example, to select constructs that have an identity of
    'air_temperature': ``d = c('air_temperature')``.

    .. versionadded:: 3.0.0

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)

    @classmethod
    def _matching_values(cls, value0, construct, value1, basic=False):
        """Whether two values match according to equality on a given
        construct.

        The definition of "match" depends on the types of *value0* and
        *value1*.

        :Parameters:

            value0:
                The first value to be matched.

            construct:
                The construct whose `_equals` method is used to
                determine whether values can be considered to match.

            value1:
                The second value to be matched.

            basic: `bool`
                If True then value0 and value1 will be compared with
                the basic ``==`` operator.

        :Returns:

            `bool`
                Whether or not the two values match.

        """
        if isinstance(value0, Query):
            return value0.evaluate(value1)

        return super()._matching_values(value0, construct, value1, basic=basic)

    def _flip(self, axes):
        """Flip (reverse the direction of) axes of the constructs in-
        place.

        .. versionadded:: 3.11.0

        :Parameters:

            axes: sequence of `str`
                Select the domain axes to flip, defined by the domain axis
                identifiers. The sequence may be empty.

        :Returns:

            `None`

        """
        data_axes = self.data_axes()

        # Flip any constructs which span the given axes
        for key, construct in self.filter_by_data().items():
            construct_axes = data_axes[key]
            construct_flip_axes = axes.intersection(construct_axes)
            if construct_flip_axes:
                iaxes = [
                    construct_axes.index(axis) for axis in construct_flip_axes
                ]
                construct.flip(iaxes, inplace=True)

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def close(self):
        """Close all files referenced by the metadata constructs.

        Note that a closed file will be automatically reopened if its
        contents are subsequently required.

        :Returns:

            `None`

        **Examples**

        >>> c.close()

        """
        # TODODASK - is this method still needed?

        for construct in self.filter_by_data().values():
            construct.close()

    def _filter_by_identity(self, arg, identities, todict, _config):
        """Worker function for `filter_by_identity` and `filter`.

        See `filter_by_identity` for details.

        .. versionadded:: 3.9.0

        """
        ctypes = [i for i in "XTYZ" if i in identities]

        config = {"identities_kwargs": {"ctypes": ctypes}}
        if _config:
            config.update(_config)

        return super()._filter_by_identity(arg, identities, todict, config)

    #    def _filter_by_coordinate_type(self, arg, ctypes, todict):
    #        """Worker function for `filter_by_identity` and `filter`.
    #
    #        See `filter_by_identity` for details.
    #
    #        .. versionadded:: 3.9.0
    #
    #        """
    #        out, pop = self._filter_preprocess(
    #            arg,
    #            filter_applied={"filter_by_identity": ctypes},
    #            todict=todict,
    #        )
    #
    #        if not ctypes:
    #            # Return all constructs if no coordinate types have been
    #            # provided
    #            return out
    #
    #        for cid, construct in tuple(out.items()):
    #            ok = False
    #            for ctype in ctypes:
    #                if getattr(construct, ctype, False):
    #                    ok = True
    #                    break
    #
    #            if not ok:
    #                pop(cid)
    #
    #        return out

    @classmethod
    def _short_iteration(cls, x):
        """The default short circuit test.

        If this method returns True then only the first identity
        returned by the construct's `!identities` method will be
        checked.

        See `_filter_by_identity` for details.

        .. versionadded:: (cfdm) 1.8.9.0

        :Parameters:

            x: `str`
                The value against which the construct's identities are
                being compared.

        :Returns:

            `bool`
                 Returns `True` if a construct's `identities` method
                 is to short circuit after the first identity is
                 computed, otherwise `False`.

        """
        if not isinstance(x, str):
            return False

        if x in "XTYZ" or x.startswith("measure:") or x.startswith("id%"):
            return True

        return "=" not in x and ":" not in x and "%" not in x



import cfdm

from . import mixin


class InteriorRing(mixin.PropertiesData, cfdm.InteriorRing):
    """An interior ring array with properties.

    If a cell is composed of multiple polygon parts, an individual
    polygon may define an "interior ring", i.e. a region that is to be
    omitted from, as opposed to included in, the cell extent. In this
    case an interior ring array is required that records whether each
    polygon is to be included or excluded from the cell, and is
    supplied by an interior ring variable in CF-netCDF. The interior
    ring array spans the same domain axis constructs as its coordinate
    array, with the addition of an extra dimension that indexes the
    parts for each cell. For example, a cell describing the land area
    surrounding a lake would require two polygon parts: one defines
    the outer boundary of the land area; the other, recorded as an
    interior ring, is the lake boundary, defining the inner boundary
    of the land area.

    **NetCDF interface**

    {{netCDF variable}}}

    The name of the netCDF dimension spanned by the interior ring
    variable's data (which does not correspond to a domain axis
    construct) may be accessed with the `nc_set_dimension`,
    `nc_get_dimension`, `nc_del_dimension` and `nc_has_dimension`
    methods.

    .. versionadded:: 3.2.0

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)



import cfdm


class RaggedContiguousArray(cfdm.RaggedContiguousArray):
    """An underlying contiguous ragged array.

    A collection of features stored using a contiguous ragged array
    combines all features along a single dimension (the "sample
    dimension") such that each feature in the collection occupies a
    contiguous block.

    The information needed to uncompress the data is stored in a
    "count variable" that gives the size of each block.

    .. versionadded:: 3.0.0

    """



import datetime
import faulthandler
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cf

# Note: it is important we test on the cf logging config rather than the
# generic Python module logging (i.e. 'cf.logging' not just 'logging').
# Also, mimic the use in the codebase by using a module-specific logger:
log_name = __name__
logger = cf.logging.getLogger(log_name)


class dummyClass:
    """Dummy class acting as container to test methods as proper
    instance methods, mirroring their context in the codebase."""

    def __init__(self, verbose=None):
        self.verbose = verbose

        self.debug_message = "A major clue to solving the evasive bug"
        self.detail_message = "In practice this will be very detailed."
        self.info_message = "This should be short and sweet"
        self.warning_message = "Best pay attention to this!"

    def func_2(self, good_kwarg=True, traceback=False, bad_kwarg=False):
        """Dummy function, otherwise trivial, where a True boolean
        passed as a traceback keyword argument will ultimately raise an
        error."""
        if traceback:
            cf.functions._DEPRECATION_ERROR_KWARGS(
                self, "another_func", traceback=True
            )
        return good_kwarg

    @cf.decorators._deprecated_kwarg_check("traceback")
    def decorated_func_2(self, good_kwarg=True, traceback=False):
        """Dummy function equivalent to 'func_2', but a decorator
        manages the logic to raise the error on use of a deprecated
        keyword argument."""
        return good_kwarg

    # Not testing 'bad_kwarg' here other than to the extent that it does not
    # stop 'traceback from causing the expected deprecation-related error.
    @cf.decorators._deprecated_kwarg_check("traceback", "bad_kwarg")
    def multikwarg_decorated_func_2(
        self, good_kwarg=True, traceback=False, bad_kwarg=False
    ):
        """Dummy function equivalent to 'func_2', but a decorator
        manages the logic to raise the error on use of a deprecated
        keyword argument."""
        return good_kwarg

    @cf.decorators._manage_log_level_via_verbose_attr
    def decorated_logging_func(self):
        """Dummy method to test _manage_log_level_via_verbose_attr.

        In particular, to test it interfaces with self.verbose
        correctly.

        """
        logger.debug(self.debug_message)
        logger.detail(self.detail_message)
        logger.info(self.info_message)
        logger.warning(self.warning_message)


class DecoratorsTest(unittest.TestCase):
    """Test decorators module.

    These are unit tests on the self-contained decorators applied to an
    artificial, trivial & not cf-python specific class, so for the cases
    where decorators are imported directly from cf, there is no need to
    duplicate such tests which are already in the cf test suite.

    """

    def setUp(self):
        self.test_only = []

    def test_deprecated_kwarg_check(self):
        test_class = dummyClass()

        # Test without (or with default) deprecated keyword argument
        res_1 = test_class.func_2(good_kwarg="good")
        res_2 = test_class.decorated_func_2(good_kwarg="good")
        res_3 = test_class.func_2(good_kwarg="good", traceback=False)
        res_4 = test_class.decorated_func_2(good_kwarg="good", traceback=False)
        test_class.multikwarg_decorated_func_2(
            good_kwarg="good", traceback=False
        )
        self.assertEqual(res_1, res_2)
        self.assertEqual(res_2, "good")
        self.assertEqual(res_3, res_4)
        self.assertEqual(res_4, "good")

        # Test with deprecated keyword argument
        with self.assertRaises(cf.functions.DeprecationError):
            test_class.func_2(good_kwarg="good", traceback=True)
        with self.assertRaises(cf.functions.DeprecationError):
            test_class.decorated_func_2(good_kwarg="good", traceback=True)
        with self.assertRaises(cf.functions.DeprecationError):
            test_class.func_2(traceback=True, bad_kwarg="bad")
        with self.assertRaises(cf.functions.DeprecationError):
            test_class.multikwarg_decorated_func_2(
                traceback=True, bad_kwarg="bad"
            )

    def test_manage_log_level_via_verbose_attr(self):
        # Order of decreasing severity/verbosity is crucial to one test below
        levels = ["WARNING", "INFO", "DETAIL", "DEBUG"]

        # Note we test assertions on the root logger object, which is the
        # one output overall at runtime, but the specific module logger name
        # should be registered within the log message:
        example_class = dummyClass()
        log_message = [
            f"WARNING:{log_name}:{example_class.warning_message}",
            f"INFO:{log_name}:{example_class.info_message}",
            f"DETAIL:{log_name}:{example_class.detail_message}",
            f"DEBUG:{log_name}:{example_class.debug_message}",
        ]

        for level in levels:
            # Important! Need to initialise class inside this loop not
            # outside it or, it retains the verbosity attribute value set
            # for the previous loop (0, i.e disable, so nothing emerges!)
            test_class = dummyClass()
            cf.log_level(level)  # reset to level

            # Default verbose(=None) cases: log_level should determine output
            with self.assertLogs(level=cf.log_level().value) as catch:
                test_class.decorated_logging_func()

                for msg in log_message:
                    # log_level should prevent messages less severe appearing:
                    if levels.index(level) >= log_message.index(msg):
                        self.assertIn(msg, catch.output)
                    else:  # less severe, should be effectively filtered out
                        self.assertNotIn(msg, catch.output)

            # Cases where verbose is set; value should override log_level...

            # Highest verbosity case (note -1 == 'DEBUG', highest verbosity):
            # all messages should appear, regardless of global log_level:
            for attr in (-1, "DEBUG", "debug", "Debug", "DeBuG"):
                test_class.verbose = attr
                with self.assertLogs(level=cf.log_level().value) as catch:
                    test_class.decorated_logging_func()
                    for msg in log_message:
                        self.assertIn(msg, catch.output)

            # Lowest verbosity case ('WARNING' / 1) excluding special case of
            # 'DISABLE' (see note above): only warning messages should appear,
            # regardless of global log_level value set:
            for attr in (1, "WARNING", "warning", "Warning", "WaRning"):
                test_class.verbose = attr
                with self.assertLogs(level=cf.log_level().value) as catch:
                    test_class.decorated_logging_func()
                    for msg in log_message:
                        if msg.split(":")[0] == "WARNING":
                            self.assertIn(msg, catch.output)
                        else:
                            self.assertNotIn(msg, catch.output)

            # Boolean cases for testing backwards compatibility...

            # ... verbose=2 should be equivalent to verbose=3 now:
            test_class.verbose = True
            with self.assertLogs(level=cf.log_level().value) as catch:
                test_class.decorated_logging_func()
                for msg in log_message:
                    if msg.split(":")[0] == "DEBUG":
                        self.assertNotIn(msg, catch.output)
                    else:
                        self.assertIn(msg, catch.output)

            # ... verbose=0 should be equivalent to verbose=0 now, so
            # test along with 'DISABLE' special case below...

            # Special 'DISABLE' (0) case: note this needs to be last as we
            # reset the log_level to it but need to use 'NOTSET' for the
            # assertLogs level, which sends all log messages through:
            for attr in (0, "DISABLE", "disable", "Disable", "DisAblE"):
                test_class.verbose = attr
                with self.assertLogs(level="NOTSET") as catch:
                    # Note: get 'AssertionError' if don't log anything at all,
                    # so to avoid this and allow check for disabled logging,
                    # first log something then disable and check that no other
                    # messages emerge:
                    logger.info(
                        "Purely to keep 'assertLog' happy: see comment!"
                    )
                    cf.log_level("DISABLE")
                    test_class.decorated_logging_func()
                    for msg in log_message:  # nothing else should be logged
                        self.assertNotIn(msg, catch.output)

            # verbose=0 should be equivalent in behaviour to verbose=0
            test_class.verbose = False
            with self.assertLogs(level="NOTSET") as catch:
                logger.info("Purely to keep 'assertLog' happy: see previous!")
                test_class.decorated_logging_func()
                for msg in log_message:  # nothing else should be logged
                    self.assertNotIn(msg, catch.output)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import faulthandler
import os
import unittest

import numpy

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class create_fieldTest(unittest.TestCase):
    filename = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
    )

    def test_create_field(self):
        # Dimension coordinates
        dim1 = cf.DimensionCoordinate(
            data=cf.Data(numpy.arange(10.0), "degrees")
        )
        dim1.standard_name = "grid_latitude"

        dim0 = cf.DimensionCoordinate(
            data=cf.Data(numpy.arange(9.0) + 20, "degrees")
        )
        dim0.standard_name = "grid_longitude"
        dim0.data[-1] += 5
        bounds = cf.Data(
            numpy.array(
                [dim0.data.array - 0.5, dim0.data.array + 0.5]
            ).transpose((1, 0))
        )
        bounds[-2, 1] = 30
        bounds[-1, :] = [30, 36]
        dim0.set_bounds(cf.Bounds(data=bounds))

        dim2 = cf.DimensionCoordinate(
            data=cf.Data([1.5]), bounds=cf.Bounds(data=cf.Data([[1, 2.0]]))
        )
        dim2.standard_name = "atmosphere_hybrid_height_coordinate"

        # Auxiliary coordinates
        ak = cf.DomainAncillary(data=cf.Data([10.0], "m"))
        ak.id = "atmosphere_hybrid_height_coordinate_ak"
        bounds = cf.Bounds(data=cf.Data([[5, 15.0]], units=ak.Units))
        ak.set_bounds(bounds)

        bk = cf.DomainAncillary(data=cf.Data([20.0]))
        bk.id = "atmosphere_hybrid_height_coordinate_bk"
        bounds = cf.Bounds(data=cf.Data([[14, 26.0]]))
        bk.set_bounds(bounds)

        aux2 = cf.AuxiliaryCoordinate(
            data=cf.Data(
                numpy.arange(-45, 45, dtype="int32").reshape(10, 9),
                units="degree_N",
            )
        )
        aux2.standard_name = "latitude"

        aux3 = cf.AuxiliaryCoordinate(
            data=cf.Data(
                numpy.arange(60, 150, dtype="int32").reshape(9, 10),
                units="degreesE",
            )
        )
        aux3.standard_name = "longitude"

        aux4 = cf.AuxiliaryCoordinate(
            data=cf.Data(
                numpy.array(
                    [
                        "alpha",
                        "beta",
                        "gamma",
                        "delta",
                        "epsilon",
                        "zeta",
                        "eta",
                        "theta",
                        "iota",
                        "kappa",
                    ],
                    dtype="S",
                )
            )
        )
        aux4.standard_name = "greek_letters"
        aux4[0] = cf.masked

        # Cell measures
        msr0 = cf.CellMeasure(
            data=cf.Data(1 + numpy.arange(90.0).reshape(9, 10) * 1234, "km 2")
        )
        msr0.measure = "area"

        # Data
        data = cf.Data(numpy.arange(90.0).reshape(10, 9), "m s-1")

        properties = {"standard_name": "eastward_wind"}

        f = cf.Field(properties=properties)

        axisX = f.set_construct(cf.DomainAxis(9))
        axisY = f.set_construct(cf.DomainAxis(10))
        axisZ = f.set_construct(cf.DomainAxis(1))

        f.set_data(data)

        x = f.set_construct(dim0)
        y = f.set_construct(dim1, axes=[axisY])
        z = f.set_construct(dim2, axes=[axisZ])

        lat = f.set_construct(aux2)
        lon = f.set_construct(aux3, axes=["X", axisY])
        f.set_construct(aux4, axes=["Y"])

        ak = f.set_construct(ak, axes=["Z"])
        bk = f.set_construct(bk, axes=[axisZ])

        # Coordinate references
        coordinate_conversion = cf.CoordinateConversion(
            parameters={
                "grid_mapping_name": "rotated_latitude_longitude",
                "grid_north_pole_latitude": 38.0,
                "grid_north_pole_longitude": 190.0,
            }
        )
        ref0 = cf.CoordinateReference(
            coordinate_conversion=coordinate_conversion,
            coordinates=[x, y, lat, lon],
        )

        f.set_construct(msr0, axes=[axisX, "Y"])

        f.set_construct(ref0)

        orog = cf.DomainAncillary()
        orog.standard_name = "surface_altitude"
        orog.set_data(cf.Data(f.array * 2, "m"))
        orog.transpose([1, 0], inplace=True)

        orog_key = f.set_construct(orog, axes=["X", axisY])

        coordinate_conversion = cf.CoordinateConversion(
            parameters={
                "standard_name": "atmosphere_hybrid_height_coordinate"
            },
            domain_ancillaries={"orog": orog_key, "a": ak, "b": bk},
        )
        ref1 = cf.CoordinateReference(
            coordinate_conversion=coordinate_conversion, coordinates=[z]
        )

        f.set_construct(ref1)

        # Field ancillary variables
        g = cf.FieldAncillary()
        g.set_data(f.data)
        g.transpose([1, 0], inplace=True)
        g.standard_name = "ancillary0"
        g *= 0.01
        f.set_construct(g)

        g = cf.FieldAncillary()
        g.set_data(f.data)
        g.standard_name = "ancillary1"
        g *= 0.01
        f.set_construct(g)

        g = cf.FieldAncillary()
        g.set_data(f[0, :].data)
        g.squeeze(inplace=True)
        g.standard_name = "ancillary2"
        g *= 0.001
        f.set_construct(g)

        g = cf.FieldAncillary()
        g.set_data(f[:, 0].data)
        g.squeeze(inplace=True)
        g.standard_name = "ancillary3"
        g *= 0.001
        f.set_construct(g)

        f.flag_values = [1, 2, 4]
        f.flag_meanings = ["a", "bb", "ccc"]

        for cm in cf.CellMethod.create(
            "grid_longitude: mean grid_latitude: max"
        ):
            f.set_construct(cm)

        # Write the file, and read it in
        cf.write(f, self.filename, verbose=0, string=True)

        g = cf.read(self.filename, squeeze=True, verbose=0)[0]

        self.assertTrue(
            g.equals(f, verbose=0), "Field not equal to itself read back in"
        )

        x = g.dump(display=False)
        x = f.dump(display=False)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



from copy import deepcopy

import numpy
from numpy import empty as numpy_empty
from numpy import expand_dims as numpy_expand_dims
from numpy import ndenumerate as numpy_ndenumerate
from numpy import squeeze as numpy_squeeze

from ..decorators import _inplace_enabled, _inplace_enabled_define_and_cleanup
from ..functions import _DEPRECATION_ERROR_METHOD
from .partition import Partition

_empty_matrix = numpy_empty((), dtype=object)


class PartitionMatrix:
    """A hyperrectangular partition matrix of a master data array.

    Each of elements (called partitions) span all or part of exactly one
    sub-array of the master data array.

    Normal numpy basic and advanced indexing is supported, but size 1
    dimensions are always removed from the output array, i.e. a partition
    rather than a partition matrix is returned if the output array has
    size 1.

    Normal numpy basic and advanced indexing is supported, but size 1
    dimensions are always removed from the output array, i.e. a partition
    rather than a partition matrix is returned if the output array has
    size 1.

    **Attributes**

    ==========  ===========================================================
    Attribute   Description
    ==========  ===========================================================
    `!axes`
    `!matrix`
    `!ndim`     The number of partition dimensions in the partition matrix.
    `!shape`    List of the partition matrix's dimension sizes.
    `!size`     The number of partitions in the partition matrix.
    ==========  ===========================================================

    """

    def __init__(self, matrix, axes):
        """**Initialization**

        :Parameters:

            matrix: `numpy.ndarray`
                An array of Partition objects.

            axes: `list`
                The identities of the partition axes of the partition
                array. If the partition matrix is a scalar array then it
                is an empty list. DO NOT UPDATE INPLACE.

        **Examples:**

        >>> pm = PartitionMatrix(
        ...          numpy.array(Partition(
        ...              location = [(0, 1), (2, 4)],
        ...              shape = [1, 2],
        ...              _dimensions = ['dim2', 'dim0'],
        ...              Units = cf.Units('m'),
        ...              part = [],
        ...              data = numpy.array([[5, 6], [7, 8]])
        ...          ), dtype=object),
        ...          axes=[]
        ...      )

        """
        self.matrix = matrix
        self.axes = axes

    def __deepcopy__(self, memo):
        """Used if copy.deepcopy is called on the variable."""
        return self.copy()

    def __getitem__(self, indices):
        """x.__getitem__(indices) <==> x[indices]

        Normal numpy basic and advanced indexing is supported, but size 1
        dimensions are always removed from the output array, i.e. a
        partition rather than a partition matrix is returned if the output
        array has size 1.

        Returns either a partition or a partition matrix.

        **Examples:**

        >>> pm.shape
        (5, 3)
        >>> pm[0, 1]
        <cf.data.partition.Partition at 0x1934c80>
        >>> pm[:, 1]
        <CF PartitionMatrix: 1 partition dimensions>
        >>> pm[:, 1].shape
        (5,)
        >>> pm[1:4, slice(2, 0, -1)].shape
        (3, 2)

        >>> pm.shape
        ()
        >>> pm[()]
        <cf.data.partition.Partition at 0x1934c80>
        >>> pm[...]
        <cf.data.partition.Partition at 0x1934c80>

        """
        out = self.matrix[indices]

        if isinstance(out, Partition):
            return out

        if out.size == 1:
            return self.matrix.item()

        axes = [axis for axis, n in zip(self.axes, out.shape) if n != 1]

        return type(self)(numpy_squeeze(out), axes)

    def __repr__(self):
        """x.__repr__() <==> repr(x)"""
        return "<CF %s: %s>" % (self.__class__.__name__, self.shape)

    def __setitem__(self, indices, value):
        """x.__setitem__(indices, y) <==> x[indices]=y.

        Indices must be an integer, a slice object or a tuple. If a slice
        object is given then the value being assigned must be an
        iterable. If a tuple of integers (or slices equivalent to an
        integer) is given then there must be one index per partition
        matrix dimension.

        **Examples:**

        >>> pm.shape
        (3,)
        >>> pm[2] = p1
        >>> pm[:] = [p1, p2, p3]

        >>> pm.shape
        (2, 3)
        >>> pm[0, 2] = p1


        >>> pm.shape
        ()
        >>> pm[()] = p1
        >>> pm[...] = p1

        """
        self.matrix[indices] = value

    def __str__(self):
        """x.__str__() <==> str(x)"""
        return str(self.matrix)
        out = []
        for partition in self.matrix.flat:
            out.append(str(partition))

        return "\n".join(out)

    def change_axis_names(self, axis_map):
        """Change the axis names.

        The axis names are arbitrary, so mapping them to another arbitrary
        collection does not change the data array values, units, nor axis
        order.

        :Parameters:

            axis_map: `dict`

        :Returns:

            `None`

        """
        # Partition dimensions
        axes = self.axes
        self.axes = [axis_map[axis] for axis in axes]

        # Partitions. Note that a partition may have dimensions which
        # are not in self.axes and that these must also be in
        # axis_name_map.
        for partition in self.matrix.flat:
            partition.change_axis_names(axis_map)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def flat(self):
        """A flat iterator over the partitions in the partition matrix.

        **Examples:**

        >>> pm.shape
        [2, 2]
        >>> for partition in pm.flat:
        ...     print(repr(partition.Units))
        ...
        <CF Units: m s-1>
        <CF Units: km hr-1>
        <CF Units: miles day-1>
        <CF Units: mm minute-1>

        >>> pm.flat
        <numpy.flatiter at 0x1e14840>

        >>> flat = pm.flat
        >>> next(flat)
        <Partition at 0x1934c80>
        >>> next(flat)
        <Partition at 0x784b347>

        """
        return self.matrix.flat

    @property
    def ndim(self):
        """The number of partition dimensions in the partition matrix.

        Not to be confused with the number of dimensions of the master
        data array.

        **Examples:**

        >>> pm.shape
        (8, 4)
        >>> pm.ndim
        2

        >>> pm.shape
        ()
        >>> pm.ndim
        0

        """
        return self.matrix.ndim

    @property
    def shape(self):
        """List of the partition matrix's dimension sizes.

        Not to be confused with the sizes of the master data array's
        dimensions.

        **Examples:**

        >>> pm.ndim
        2
        >>> pm.size
        32
        >>> pm.shape
        (8, 4)

        >>> pm.ndim
        0
        >>> pm.shape
        ()

        """
        return self.matrix.shape

    @property
    def size(self):
        """The number of partitions in the partition matrix.

        Not to be confused with the number of elements in the master data
        array.

        **Examples:**

        >>> pm.shape
        (8, 4)
        >>> pm.size
        32

        >>> pm.shape
        ()
        >>> pm.size
        1

        """
        return self.matrix.size

    def add_partitions(self, adimensions, master_flip, extra_boundaries, axis):
        """Add partition boundaries.

        :Parameters:

            adimensions: `list`
                The ordered axis names of the master array.

            master_flip: `list`

            extra_boundaries: `list` of `int`
                The boundaries of the new partitions.

            axis: `str`
                The name of the axis to have the new partitions.

        """

        def _update_p(
            matrix,
            location,
            master_index,
            part,
            master_axis_to_position,
            master_flip,
        ):
            """Create a new partition matrix via an update.

            :Parameters:

                matrix: numpy array of `Partition` objects

                location: `list`

                master_index: `int`

                part: `list`

                master_axis_to_position: `dict`

                master_flip: `list`

            :Returns:

                numpy array of `Partition` objects

            """
            for partition in matrix.flat:
                partition.location = partition.location[:]
                partition.shape = partition.shape[:]

                partition.location[master_index] = location
                partition.shape[master_index] = shape

                partition.new_part(part, master_axis_to_position, master_flip)
            # --- End: for

            return matrix

        # If no extra boundaries have been provided, just return
        # without doing anything
        if not extra_boundaries:
            return

        master_index = adimensions.index(axis)
        index = self.axes.index(axis)

        # Find the position of the extra-boundaries dimension in the
        # list of master array dimensions
        extra_boundaries = extra_boundaries[:]

        # Create the master_axis_to_position dictionary required by
        # Partition.new_part
        master_axis_to_position = {}
        for i, data_axis in enumerate(adimensions):
            master_axis_to_position[data_axis] = i

        matrix = self.matrix
        shape = matrix.shape

        # Initialize the new partition matrix
        new_shape = list(shape)
        new_shape[index] += len(extra_boundaries)
        new_matrix = numpy_empty(new_shape, dtype=object)

        part = [slice(None)] * len(adimensions)
        indices = [slice(None)] * matrix.ndim
        new_indices = indices[:]
        new_indices[index] = slice(0, 1)  # 0

        # Find the first extra boundary
        x = extra_boundaries.pop(0)

        for i in range(shape[index]):
            indices[index] = slice(i, i + 1)
            sub_matrix = matrix[tuple(indices)]
            #            (r0, r1) = next(sub_matrix.flat).location[master_index]
            (r0, r1) = sub_matrix.item(
                0,
            ).location[master_index]

            # Could do better, perhaps, by assigning in blocks
            if not r0 < x < r1:
                # This new boundary is *not* within the span of this
                # sub-matrix. Therefore, just copy the sub-matrix
                # straight into the new matrix
                new_matrix[tuple(new_indices)] = sub_matrix
                #                new_indices[index] += 1
                new_indices[index] = slice(
                    new_indices[index].start + 1, new_indices[index].stop + 1
                )
                continue

            # --------------------------------------------------------
            # Still here? Then this new boundary *is* within the span
            # of this sub-matrix.
            # --------------------------------------------------------

            # Find the new extent of the original partition(s)
            location = (r0, x)
            shape = x - r0
            part[master_index] = slice(0, shape)

            # Create new partition(s) in place of the original ones(s)
            # and set the location, shape and part attributes
            new_matrix[tuple(new_indices)] = _update_p(
                deepcopy(sub_matrix),
                location,
                master_index,
                part,
                master_axis_to_position,
                master_flip,
            )
            #            new_indices[index] += 1
            new_indices[index] = slice(
                new_indices[index].start + 1, new_indices[index].stop + 1
            )

            while x < r1:
                # Find the extent of the new partition(s)
                if not extra_boundaries:
                    # There are no more new boundaries, so the new
                    # partition(s) run to the end of the original
                    # partition(s) in which they lie.
                    location1 = r1
                else:
                    # There are more new boundaries, so this
                    # new partition runs either to the next
                    # new boundary or to the end of the
                    # original partition, which comes first.
                    location1 = min(extra_boundaries[0], r1)

                location = (x, location1)
                shape = location1 - x
                offset = x - r0
                part[master_index] = slice(offset, offset + shape)

                # Create the new partition(s) and set the
                # location, shape and part attributes
                new_matrix[tuple(new_indices)] = _update_p(
                    deepcopy(sub_matrix),
                    location,
                    master_index,
                    part,
                    master_axis_to_position,
                    master_flip,
                )

                new_indices[index] = slice(
                    new_indices[index].start + 1, new_indices[index].stop + 1
                )
                #                new_indices[index] += 1

                if not extra_boundaries:
                    # There are no more extra boundaries, so we can
                    # return now.
                    #  new_indices[index] = slice(new_indices[index],
                    #  None)
                    new_indices[index] = slice(new_indices[index].start, None)
                    indices[index] = slice(i + 1, None)

                    new_matrix[tuple(new_indices)] = matrix[tuple(indices)]
                    self.matrix = new_matrix

                    return

                # Still here? Then move on to the next new boundary
                x = extra_boundaries.pop(0)
            # --- End: while
        # --- End: for

        self.matrix = new_matrix

    def copy(self):
        """Return a deep copy.

        ``pm.copy()`` is equivalent to ``copy.deepcopy(pm)``.

        :Returns:

                The deep copy.

        **Examples:**

        >>> pm.copy()

        """
        # ------------------------------------------------------------
        # NOTE: 15 May 2013. It is necesary to treat
        #       self.matrix.ndim==0 as a special case since there is a
        #       bug (feature?) in numpy <= v1.7 (at least):
        #       http://numpy-discussion.10968.n7.nabble.com/bug-in-deepcopy-of-rank-zero-arrays-td33705.html
        # ------------------------------------------------------------
        matrix = self.matrix

        if not matrix.ndim:
            new_matrix = _empty_matrix.copy()  # numpy_empty((), dtype=object)
            new_matrix[()] = matrix.item().copy()
            return type(self)(new_matrix, [])
        else:
            new_matrix = numpy.empty(matrix.size, dtype=object)
            new_matrix[...] = [partition.copy() for partition in matrix.flat]
            new_matrix.resize(matrix.shape)
            return type(self)(new_matrix, self.axes)

    # 0
    @_inplace_enabled(default=False)
    def insert_dimension(self, axis, inplace=False):
        """Insert a new size 1 axis in place.

        The new axis is always inserted at position 0, i.e. it becomes the
        new slowest varying axis.

        .. seealso:: `flip`, `squeezes`, `swapaxes`, `transpose`

        :Parameters:

            axis: `str`
                The internal identity of the new axis.

        :Returns:

            `PartitionMatrix`

        **Examples:**

        >>> pm.shape
        (2, 3)
        >>> pm.insert_dimension('dim2')
        >>> pm.shape
        (1, 2, 3)

        """
        p = _inplace_enabled_define_and_cleanup(self)

        p.matrix = numpy_expand_dims(p.matrix, 0)
        p.axes = [axis] + p.axes

        return p

    def ndenumerate(self):
        """Return an iterator yielding pairs of array indices and
        values.

        :Returns:

            `numpy.ndenumerate`
                An iterator over the array coordinates and values.

        **Examples:**

        >>> pm.shape
        (2, 3)
        >>> for i, partition in pm.ndenumerate():
        ...     print(i, repr(partition))
        ...
        (0, 0) <cf.data.partition.Partition object at 0x13a4490>
        (0, 1) <cf.data.partition.Partition object at 0x24a4650>
        (0, 2) <cf.data.partition.Partition object at 0x35a4590>
        (1, 0) <cf.data.partition.Partition object at 0x46a4789>
        (1, 1) <cf.data.partition.Partition object at 0x57a3456>
        (1, 2) <cf.data.partition.Partition object at 0x68a9872>

        """
        return numpy_ndenumerate(self.matrix)

    # 0
    def partition_boundaries(self, data_axes):
        """Return the partition boundaries for each dimension.

        :Parameters:

            data_axes: sequence

        :Returns:

            `dict`

        """
        boundaries = {}

        matrix = self.matrix
        indices = [0] * self.ndim

        for i, axis in enumerate(self.axes):
            indices[i] = slice(None)
            j = data_axes.index(axis)

            sub_matrix = matrix[tuple(indices)]

            b = [partition.location[j][0] for partition in sub_matrix.flat]

            # Python3: can't access variables from within previous
            # list comprehension
            last_partition = sub_matrix.flat[-1]

            b.append(last_partition.location[j][1])
            boundaries[axis] = b

            indices[i] = 0
        # --- End: for

        return boundaries

    # 0
    @_inplace_enabled(default=False)
    def swapaxes(self, axis0, axis1, inplace=False):
        """Swap the positions of two axes.

        Note that this does not change the master data array.

        .. seealso:: `insert_dimension`, `flip`, `squeeze`, `transpose`

        :Parameters:

            axis0, axis1: `int`, `int`
                Select the axes to swap. Each axis is identified by its
                original integer position.

        :Returns:

            `PartitionMatrix`

        **Examples:**

        >>> pm.shape
        (2, 3, 4, 5)
        >>> pm.swapaxes(1, 2)
        >>> pm.shape
        (2, 4, 3, 5)
        >>> pm.swapaxes(1, -1)
        >>> pm.shape
        (2, 5, 3, 4)

        """
        p = _inplace_enabled_define_and_cleanup(self)

        if axis0 != axis1:
            iaxes = list(range(p.matrix.ndim))
            iaxes[axis1], iaxes[axis0] = iaxes[axis0], iaxes[axis1]
            p.transpose(iaxes, inplace=True)

        return p

    # 0
    def set_location_map(self, data_axes, ns=None):
        """Set the `!location` attribute of each partition of the
        partition matrix in place.

        :Parameters:

            data_axes: sequence
                The axes of the master data array.

            ns: sequence of `int`, optional

        :Returns:

            `None`

        **Examples:**

        >>> pm.set_location_map(['dim1', 'dim0'])
        >>> pm.set_location_map([])

        """
        matrix = self.matrix

        shape = matrix.shape
        axes = self.axes

        slice_None = slice(None)

        indices = [slice_None] * matrix.ndim

        # Never update location in-place
        for partition in matrix.flat:
            partition.location = partition.location[:]

        if ns is None:
            ns = range(len(data_axes))

        for axis, n in zip(data_axes, ns):

            if axis in axes:
                # ----------------------------------------------------
                # This data array axis is also a partition matrix axis
                # ----------------------------------------------------
                m = axes.index(axis)
                start = 0
                for i in range(shape[m]):
                    indices[m] = slice(i, i + 1)  # i
                    flat = matrix[tuple(indices)].flat

                    partition = next(flat)
                    stop = start + partition.shape[n]
                    location = (start, stop)

                    partition.location[n] = location

                    for partition in flat:
                        partition.location[n] = location

                    start = stop

                # --- End: for
                indices[m] = slice_None
            else:
                # ----------------------------------------------------
                # This data array axis is not a partition matrix axis
                # ----------------------------------------------------
                flat = matrix.flat
                partition = next(flat)
                location = (0, partition.shape[n])

                partition.location[n] = location

                for partition in flat:
                    partition.location[n] = location
        # --- End: for

    # 0
    @_inplace_enabled(default=False)
    def squeeze(self, inplace=False):
        """Remove all size 1 axes.

        Note that this does not change the master data array.

        .. seealso:: `insert_dimension`, `flip`, `swapaxes`, `transpose`

        :Returns:

            `PartitionMatrix`

        **Examples:**

        >>> pm.shape
        (1, 2, 1, 2)
        >>> pm.squeeze()
        >>> pm.shape
        (2, 2)

        >>> pm.shape
        (1,)
        >>> pm.squeeze()
        >>> pm.shape
        ()
        >>> pm.squeeze()
        >>> pm.shape
        ()

        """
        p = _inplace_enabled_define_and_cleanup(self)

        matrix = p.matrix
        shape = matrix.shape

        if 1 in shape:
            p.matrix = matrix.squeeze()

            axes = p.axes
            p.axes = [axis for axis, size in zip(axes, shape) if size > 1]

        return p

    @_inplace_enabled(default=False)
    def transpose(self, axes, inplace=False):
        """Permute the partition dimensions of the partition matrix in
        place.

        Note that this does not change the master data array.

        .. seealso:: `insert_dimension`, `flip`, `squeeze`, `swapaxes`

        :Parameters:

            axes: sequence of `int`
                Permute the axes according to the values given.

        :Returns:

            `PartitionMatrix`

        **Examples:**

        >>> pm.ndim
        3
        >>> pm.transpose((2, 0, 1))

        """
        p = _inplace_enabled_define_and_cleanup(self)

        matrix = p.matrix
        if list(axes) != list(range(matrix.ndim)):
            p.matrix = matrix.transpose(axes)
            p_axes = p.axes
            p.axes = [p_axes[i] for i in axes]

        return p

    # ----------------------------------------------------------------
    # Deprecated attributes and methods
    # ----------------------------------------------------------------
    def expand_dims(self, axis, inplace=False):
        """Insert a new size 1 axis in place.

        Deprecated at version 3.0.0. Use method 'insert_dimension'
        instead.

        """
        _DEPRECATION_ERROR_METHOD(
            self, "expand_dims", "Use method 'insert_dimension' instead."
        )  # pragma: no cover


# --- End: class



from numpy.ma import masked_where as numpy_ma_masked_where

from ...constants import _file_to_fh
from ...data.abstract.filearray import FileArray
from ...functions import get_subspace, parse_indices
from ...umread_lib.umfile import Rec
from .functions import _close_um_file, _open_um_file

_filename_to_file = _file_to_fh.setdefault("UM", {})


class UMFileArray(FileArray):
    """A sub-array stored in a PP or UM fields file.

    **Initialization**

    :Parameters:

        file: `str`
            The file name in normalized, absolute form.

        dtype: `numpy.dtype`
            The data type of the data array on disk.

        ndim: `int`
            The number of dimensions in the unpacked data array.

        shape: `tuple`
            The shape of the unpacked data array.

        size: `int`
            The number of elements in the unpacked data array.

        header_offset: `int`
            The start position in the file of the header.

        data_offset: `int`
            The start position in the file of the data array.

        disk_length: `int`
            The number of words on disk for the data array, usually
            LBLREC-LBEXT. If set to 0 then `!size` is used.

        fmt: `str`, optional

        word_size: `int`, optional

        byte_ordering: `str`, optional

    **Examples**

    >>> a = UMFileArray(file='file.pp', header_offset=3156,
    ...                 data_offset=3420,
    ...                 dtype=numpy.dtype('float32'), shape=(30, 24),
    ...                 size=720, ndim=2, disk_length=0)

    >>> a = UMFileArray(file='packed_file.pp', header_offset=3156,
    ...                 data_offset=3420, dtype=numpy.dtype('float32'),
    ...                 shape=(30, 24), size=720, ndim=2,
    ...                 disk_length=423)

    """

    def __getitem__(self, indices):
        """Implement indexing.

        x.__getitem__(indices) <==> x[indices]

        Returns a numpy array.

        """
        f = self.open()

        rec = Rec.from_file_and_offsets(
            f, self.header_offset, self.data_offset, self.disk_length
        )

        int_hdr = rec.int_hdr
        real_hdr = rec.real_hdr

        array = rec.get_data().reshape(
            int_hdr.item(
                17,
            ),
            int_hdr.item(
                18,
            ),
        )

        if indices is not Ellipsis:
            indices = parse_indices(array.shape, indices)
            array = get_subspace(array, indices)

        LBUSER2 = int_hdr.item(
            38,
        )

        if LBUSER2 == 3:
            # Return the numpy array now if it is a boolean array
            return array.astype(bool)

        integer_array = LBUSER2 == 2

        # ------------------------------------------------------------
        # Convert to a masked array
        # ------------------------------------------------------------
        # Set the fill_value from BMDI
        fill_value = real_hdr.item(
            17,
        )
        if fill_value != -1.0e30:
            # -1.0e30 is the flag for no missing data
            if integer_array:
                # The fill_value must be of the same type as the data
                # values
                fill_value = int(fill_value)

            # Mask any missing values
            mask = array == fill_value
            if mask.any():
                array = numpy_ma_masked_where(mask, array, copy=False)

        # ------------------------------------------------------------
        # Unpack the array using the scale_factor and add_offset, if
        # either is available
        # ------------------------------------------------------------
        # Treat BMKS as a scale_factor if it is neither 0 nor 1
        scale_factor = real_hdr.item(
            18,
        )
        if scale_factor != 1.0 and scale_factor != 0.0:
            if integer_array:
                scale_factor = int(scale_factor)
            array *= scale_factor

        # Treat BDATUM as an add_offset if it is not 0
        add_offset = real_hdr.item(
            4,
        )
        if add_offset != 0.0:
            if integer_array:
                add_offset = int(add_offset)
            array += add_offset

        # Return the numpy array
        return array

    def __str__(self):
        """x.__str__() <==> str(x)"""
        return "%s%s in %s" % (self.header_offset, self.shape, self.file)

    @property
    def file_pointer(self):
        """The file pointer starting at the position of the header."""
        return (self.file, self.header_offset)

    def close(self):
        """Close the file containing the data array.

        If the file is not open then no action is taken.

        :Returns:

            `None`

        **Examples**

        >>> f.close()

        """
        _close_um_file(self.file)

    def open(self):
        """Open the file containing the data array.

        :Returns:

            `um.umread.umfile.File`

        **Examples**

        >>> f.open()

        """
        return _open_um_file(
            self.file,
            fmt=getattr(self, "fmt", None),
            word_size=getattr(self, "word_size", None),
            byte_ordering=getattr(self, "byte_ordering", None),
        )



import datetime
import faulthandler
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class CellMeasureTest(unittest.TestCase):
    f = cf.example_field(1)

    def test_CellMeasure__repr__str__dump(self):
        x = self.f.cell_measure("measure:area")

        repr(x)
        str(x)
        x.dump(display=False)

    def test_CellMeasure_measure(self):
        x = self.f.cell_measure("measure:area").copy()

        self.assertEqual(x.measure, "area")
        del x.measure
        self.assertIsNone(getattr(x, "measure", None))
        x.measure = "qwerty"
        self.assertEqual(x.measure, "qwerty")

    def test_CellMeasure_identity(self):
        x = self.f.cell_measure("measure:area").copy()

        self.assertEqual(x.identity(), "measure:area")
        del x.measure
        self.assertEqual(x.identity(), "ncvar%cell_measure", x.identity())
        x.nc_del_variable()
        self.assertEqual(x.identity(), "")


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import faulthandler
import json
import os
import stat
import subprocess
import unittest

import netCDF4

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class cfaTest(unittest.TestCase):
    def setUp(self):
        self.test_file = "cfa_test.sh"
        self.test_path = os.path.join(os.getcwd(), self.test_file)

        # Need ./cfa_test.sh to be made executable to run it. Locally that
        # may already be true but for testing dists etc. must chmod here:
        os.chmod(
            self.test_path,
            stat.S_IRUSR
            | stat.S_IROTH
            | stat.S_IRGRP
            | stat.S_IXUSR  # reading
            | stat.S_IXOTH
            | stat.S_IXGRP  # executing
            | stat.S_IWUSR,  # writing
        )

    def test_cfa(self):
        # In the script, STDERR from cfa commands is redirected to
        # (overwrite) its STDOUT, so Popen's stdout is really the cfa
        # commands' stderr:
        cfa_test = subprocess.Popen(
            ["./" + self.test_file],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        cfa_stderr_via_stdout_channel, _ = cfa_test.communicate("yes\n")
        returncode = cfa_test.returncode
        if returncode != 0:
            self.fail(
                f"A cfa command failed (see script's 'exit {returncode}' "
                "point) with error:\n"
                f"{cfa_stderr_via_stdout_channel.decode('utf-8')}"
            )
        # else: (passes by default)

    def test_cfa_base(self):
        # Test the "base" cfa_option to cf.write. Valid for CFA-0.4.
        filename = "test_file.nc"
        f = cf.read(filename)
        cfa_file = "test_file.nca"
        cf.write(f, cfa_file, fmt="CFA4", cfa_options={"base": ""})

        nc = netCDF4.Dataset(cfa_file, "r")
        cfa_array = json.loads(nc.variables["eastward_wind"].cfa_array)

        self.assertEqual(cfa_array["base"], "")
        self.assertEqual(
            cfa_array["Partitions"][0]["subarray"]["file"], filename
        )

        os.remove(cfa_file)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import faulthandler
import os
import platform
import sys
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class functionTest(unittest.TestCase):
    def setUp(self):
        self.test_only = ()

    def test_example_field(self):
        for f in cf.example_fields():
            f.dump(display=False)

        with self.assertRaises(ValueError):
            cf.example_field(-999)

    def test_keyword_deprecation(self):
        # Use as test case 'i' kwarg, the deprecated old name for
        # 'inplace':
        a = cf.Data([list(range(100))])
        a.squeeze(inplace=True)  # new way to specify operation tested below

        b = cf.Data([list(range(100))])
        with self.assertRaises(cf.functions.DeprecationError):
            b.squeeze(i=True)

    def test_aliases(self):
        self.assertEqual(cf.log_level(), cf.LOG_LEVEL())
        self.assertEqual(cf.free_memory(), cf.FREE_MEMORY())
        self.assertEqual(cf.free_memory_factor(), cf.FREE_MEMORY_FACTOR())
        self.assertEqual(cf.fm_threshold(), cf.FM_THRESHOLD())
        self.assertEqual(cf.total_memory(), cf.TOTAL_MEMORY())
        self.assertEqual(cf.regrid_logging(), cf.REGRID_LOGGING())
        self.assertEqual(cf.relaxed_identities(), cf.RELAXED_IDENTITIES())
        self.assertEqual(cf.tempdir(), cf.TEMPDIR())
        self.assertEqual(cf.chunksize(), cf.CHUNKSIZE())
        self.assertEqual(cf.set_performance(), cf.SET_PERFORMANCE())
        self.assertEqual(cf.of_fraction(), cf.OF_FRACTION())
        self.assertEqual(
            cf.collapse_parallel_mode(), cf.COLLAPSE_PARALLEL_MODE()
        )

    def test_configuration(self):
        # This test assumes 'total_memory' remains constant throughout
        # the test run, which should be true generally in any
        # reasonable context.

        # Test getting of all config. and store original values to
        # test on:
        org = cf.configuration()
        self.assertIsInstance(org, dict)

        # Check all keys that should be there are, with correct value type:
        self.assertEqual(len(org), 11)  # update expected len if add new key(s)

        # Floats expected as values for most keys. Store these for
        # later as floats need assertAlmostEqual rather than
        # assertEqual tests:
        keys_with_float_values = [
            "atol",
            "rtol",
            "of_fraction",
            "free_memory_factor",
            "chunksize",
        ]
        for key in keys_with_float_values:
            self.assertIsInstance(org[key], float)

        # Other types expected:
        self.assertIsInstance(org["collapse_parallel_mode"], int)
        self.assertIsInstance(org["relaxed_identities"], bool)
        self.assertIsInstance(org["bounds_combination_mode"], str)
        self.assertIsInstance(org["regrid_logging"], bool)
        # Log level may be input as an int but always given as
        # equiv. string
        self.assertIsInstance(org["log_level"], str)
        self.assertIsInstance(org["tempdir"], str)

        # Store some sensible values to reset items to for testing, ensuring:
        # 1) they are kept different to the defaults (i.e. org values); and
        # 2) floats differ sufficiently that they will be picked up as
        #    qdifferent by the assertAlmostEqual decimal places (8, see
        #    below)
        reset_values = {
            "rtol": 5e-7,
            "atol": 2e-7,
            "tempdir": "/my-custom-tmpdir",
            "of_fraction": 0.1,
            "free_memory_factor": 0.25,
            "regrid_logging": True,
            "collapse_parallel_mode": 2,
            "relaxed_identities": True,
            "bounds_combination_mode": "XOR",
            "log_level": "INFO",
            "chunksize": 8e9,
        }

        # Test the setting of each lone item.
        expected_post_set = dict(org)  # copy for safety with mutable dict
        for setting, value in reset_values.items():
            cf.configuration(**{setting: value})
            post_set = cf.configuration()

            # Expect a dict that is identical to the original to start
            # with but as we set values incrementally they should be
            # reflected:
            expected_post_set[setting] = value

            # Can't trivially do a direct test that the actual and
            # expected return dicts are the same as there are float
            # values which have limited float precision so need
            # assertAlmostEqual testing:
            for name, val in expected_post_set.items():
                if isinstance(val, float):
                    self.assertAlmostEqual(
                        post_set[name], val, places=8, msg=setting
                    )
                else:
                    self.assertEqual(post_set[name], val)

        # Test the setting of more than one, but not all, items
        # simultaneously:
        new_values = {
            "regrid_logging": True,
            "tempdir": "/bin/bag",
            "of_fraction": 0.33,
        }
        cf.configuration(**new_values)
        post_set = cf.configuration()
        for name, val in new_values.items():  # test values that should change
            self.assertEqual(post_set[name], val)
        # ...and some values that should not:
        self.assertEqual(post_set["log_level"], "INFO")
        self.assertAlmostEqual(post_set["rtol"], 5e-7)

        # Test setting all possible items simultaneously (back to originals):
        cf.configuration(**org)
        post_set = cf.configuration()
        for name, val in org.items():
            if isinstance(val, float):
                self.assertAlmostEqual(post_set[name], val, places=8)
            else:
                self.assertEqual(post_set[name], val)

        # Test edge cases & invalid inputs...
        # ... 1. Falsy value inputs on some representative items:
        pre_set_config = cf.configuration()
        with self.assertRaises(ValueError):
            cf.configuration(of_fraction=0.0)
        with self.assertRaises(ValueError):
            cf.configuration(free_memory_factor=0.0)
        new_values = {
            "tempdir": "",
            "atol": 0.0,
            "regrid_logging": False,
        }
        cf.configuration(**new_values)
        post_set = cf.configuration()
        for name, val in new_values.items():  # test values that should change
            self.assertEqual(post_set[name], val)
        # ...and some values that should not:
        self.assertEqual(post_set["log_level"], pre_set_config["log_level"])
        self.assertAlmostEqual(post_set["rtol"], pre_set_config["rtol"])

        # 2. None as an input kwarg rather than as a default:
        pre_set_config = cf.configuration()
        set_of = 0.45
        cf.configuration(of_fraction=set_of, rtol=None, log_level=None)
        post_set = cf.configuration()
        # test values that should change
        self.assertEqual(post_set["of_fraction"], set_of)
        # ...and values that should not:
        self.assertEqual(post_set["rtol"], pre_set_config["rtol"])
        self.assertAlmostEqual(
            post_set["log_level"], pre_set_config["log_level"]
        )

        # 3. Gracefully error with invalid inputs:
        with self.assertRaises(ValueError):
            cf.configuration(of_fraction="bad")

        with self.assertRaises(ValueError):
            cf.configuration(log_level=7)

        # 4. Check invalid kwarg given logic processes **kwargs:
        with self.assertRaises(TypeError):
            cf.configuration(bad_kwarg=1e-15)

        old = cf.configuration()
        try:
            cf.configuration(atol=888, rtol=999, log_level="BAD")
        except ValueError:
            self.assertEqual(cf.configuration(), old)
        else:
            raise RuntimeError(
                "A ValueError should have been raised, but wasn't"
            )

        # Reset so later test fixtures don't spam with output
        # messages:
        cf.log_level("DISABLE")

    def test_context_managers(self):
        # rtol, atol, chunksize
        for func in (
            cf.atol,
            cf.rtol,
            cf.chunksize,
            cf.free_memory_factor,
            cf.of_fraction,
        ):
            old = func()
            new = old * 1.001
            with func(new):
                self.assertEqual(func(), new)
                self.assertEqual(func(new * 1.001), new)
                self.assertEqual(func(), new * 1.001)

            self.assertEqual(func(), old)

        # collapse_parallel_mode
        func = cf.collapse_parallel_mode

        org = func(0)
        old = func()
        new = old + 1
        with func(new):
            self.assertEqual(func(), new)
            self.assertEqual(func(new + 1), new)
            self.assertEqual(func(), new + 1)

        self.assertEqual(func(), old)
        func(org)

        # log_level
        func = cf.log_level

        org = func("DETAIL")
        old = func()
        new = "DEBUG"
        with func(new):
            self.assertEqual(func(), new)

        self.assertEqual(func(), old)
        func(org)

        del org._func
        with self.assertRaises(AttributeError):
            with org:
                pass

        # bounds_combination_mode
        func = cf.bounds_combination_mode

        org = func("XOR")
        old = func()
        new = "AND"
        with func(new):
            self.assertEqual(func(), new)

        self.assertEqual(func(), old)
        func(org)

        del org._func
        with self.assertRaises(AttributeError):
            with org:
                pass

        # Full configuration
        func = cf.configuration

        org = func(rtol=cf.Constant(10), atol=20, log_level="WARNING")
        old = func()
        new = dict(rtol=cf.Constant(20), atol=40, log_level="DISABLE")

        with func(**new):
            self.assertEqual(cf.atol(), 40)

        self.assertEqual(func(), old)
        func(**org)

    def test_Constant(self):
        c = cf.atol()
        self.assertIs(c._func, cf.atol)

    def test_Configuration(self):
        c = cf.Configuration()
        self.assertIs(c._func, cf.configuration)

    def test_environment(self):
        e = cf.environment(display=False)
        ep = cf.environment(display=False, paths=False)
        self.assertIsInstance(e, str)
        self.assertIsInstance(ep, str)

        components = ["Platform: ", "udunits2 library: ", "numpy: ", "cfdm: "]
        for component in components:
            self.assertIn(component, e)
            self.assertIn(component, ep)
        for component in [
            f"cf: {cf.__version__} {os.path.abspath(cf.__file__)}",
            f"Python: {platform.python_version()} {sys.executable}",
        ]:
            self.assertIn(component, e)
            self.assertNotIn(component, ep)  # paths shouldn't be present here
        for component in [
            f"cf: {cf.__version__}",
            f"Python: {platform.python_version()}",
        ]:
            self.assertIn(component, ep)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



from os import close
from tempfile import mkdtemp, mkstemp

from numpy import load as numpy_load
from numpy import ndarray as numpy_ndarray
from numpy import save as numpy_save
from numpy.ma import array as numpy_ma_array
from numpy.ma import is_masked as numpy_ma_is_masked

from ..constants import CONSTANTS
from ..functions import get_subspace, parse_indices
from . import abstract


class CachedArray(abstract.FileArray):
    """A indexable N-dimensional array supporting masked values.

    The array is stored on disk in a temporary file until it is
    accessed. The directory containing the temporary file may be found
    and set with the `cf.tempdir` function.

    """

    def __init__(self, array):
        """**Initialization**

        :Parameters:

            array: numpy array
                The array to be stored on disk in a temporary file.

        **Examples:**

        >>> f = CachedArray(numpy.array([1, 2, 3, 4, 5]))
        >>> f = CachedArray(numpy.ma.array([1, 2, 3, 4, 5]))

        """
        super().__init__()

        # ------------------------------------------------------------
        # Use mkstemp because we want to be responsible for deleting
        # the temporary file when done with it.
        # ------------------------------------------------------------
        _partition_dir = mkdtemp(
            prefix="cf_cachedarray_", dir=CONSTANTS["TEMPDIR"]
        )
        fd, _partition_file = mkstemp(
            prefix="cf_cachedarray_", suffix=".npy", dir=_partition_dir
        )
        close(fd)

        # The name of the temporary file storing the array
        self._set_component("_partition_dir", _partition_dir, copy=False)
        self._set_component("_partition_file", _partition_file, copy=False)

        # Numpy data type of the array
        self._set_component("dtype", array.dtype, copy=False)

        # Tuple of the array's dimension sizes
        self._set_component("shape", array.shape, copy=False)

        # Number of elements in the array
        self._set_component("size", array.size, copy=False)

        # Number of dimensions in the array
        self._set_component("ndim", array.ndim, copy=False)

        if numpy_ma_is_masked(array):
            # Array is a masked array. Save it as record array with
            # 'data' and 'mask' elements because this seems much
            # faster than using numpy.ma.dump.
            self._set_component("_masked_as_record", True, copy=False)
            numpy_save(_partition_file, array.toflex())
        else:
            self._set_component("_masked_as_record", False, copy=False)
            if hasattr(array, "mask"):
                # Array is a masked array with no masked elements
                numpy_save(_partition_file, array.view(numpy_ndarray))
            else:
                # Array is not a masked array.
                numpy_save(_partition_file, array)

    def __getitem__(self, indices):
        """x.__getitem__(indices) <==> x[indices]

        Returns a numpy array.

        """
        array = numpy_load(self._partition_file)

        indices = parse_indices(array.shape, indices)

        array = get_subspace(array, indices)

        if self._get_component("_masked_as_record"):
            # Convert a record array to a masked array
            array = numpy_ma_array(
                array["_data"], mask=array["_mask"], copy=False
            )
            array.shrink_mask()

        # Return the numpy array
        return array

    def __str__(self):
        """x.__str__() <==> str(x)"""
        return "%s in %s" % (self.shape, self._partition_file)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def _partition_dir(self):
        """The name of the directory containing the file storing the
        array."""
        return self._get_component("_partition_dir")

    @property
    def _partition_file(self):
        """The name of the file storing the array on disk."""
        return self._get_component("_partition_file")



import datetime
import faulthandler
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class FieldAncillaryTest(unittest.TestCase):
    f = cf.example_field(1)

    def test_FieldAncillary(self):
        f = cf.FieldAncillary()

        repr(f)
        str(f)
        f.dump(display=False)

    def test_FieldAncillary_source(self):
        a = self.f.auxiliary_coordinate("latitude")
        cf.FieldAncillary(source=a)

    def test_FieldAncillary_properties(self):
        d = self.f.domain_ancillary("ncvar%a")
        x = cf.FieldAncillary(source=d)

        x.set_property("long_name", "qwerty")

        self.assertEqual(x.get_property("long_name"), "qwerty")
        self.assertEqual(x.del_property("long_name"), "qwerty")
        self.assertIsNone(x.get_property("long_name", None))
        self.assertIsNone(x.del_property("long_name", None))

    def test_FieldAncillary_insert_dimension(self):
        d = self.f.dimension_coordinate("grid_longitude")
        x = cf.FieldAncillary(source=d)

        self.assertEqual(x.shape, (9,))

        y = x.insert_dimension(0)
        self.assertEqual(y.shape, (1, 9))

        x.insert_dimension(-1, inplace=True)
        self.assertEqual(x.shape, (9, 1))

    def test_FieldAncillary_transpose(self):
        a = self.f.auxiliary_coordinate("longitude")
        x = cf.FieldAncillary(source=a)

        self.assertEqual(x.shape, (9, 10))

        y = x.transpose()
        self.assertEqual(y.shape, (10, 9))

        x.transpose([1, 0], inplace=True)
        self.assertEqual(x.shape, (10, 9))

    def test_FieldAncillary_squeeze(self):
        a = self.f.auxiliary_coordinate("longitude")
        x = cf.FieldAncillary(source=a)

        x.insert_dimension(1, inplace=True)
        x.insert_dimension(0, inplace=True)

        self.assertEqual(x.shape, (1, 9, 1, 10))

        y = x.squeeze()
        self.assertEqual(y.shape, (9, 10))

        x.squeeze(2, inplace=True)
        self.assertEqual(x.shape, (1, 9, 10))


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



from os.path import isfile

from netCDF4 import Dataset as netCDF4_Dataset

from ..constants import _file_to_fh
from ..functions import close_one_file, open_files_threshold_exceeded
from ..umread_lib.umfile import File

_file_to_UM = _file_to_fh.setdefault("UM", {})
_file_to_Dataset = _file_to_fh.setdefault("netCDF", {})


def _open_netcdf_file(filename, mode, fmt="NETCDF4"):  # set_auto_mask=True):
    """Open a netCDF file and read it into a netCDF4.Dataset object.

    If the file is already open then the existing netCDF4.Dataset
    object will be returned.

    :Parameters:

        filename: `str`
            The netCDF file to be opened.

    #    set_auto_mask: `bool`, optional
    #        Turn on or off automatic conversion of variable data to and
    #        from masked arrays.

    :Returns:

        `netCDF4.Dataset`
            A netCDF4.Dataset instance for the netCDF file.

    **Examples:**

    >>> nc1 = _open_netcdf_file('file.nc')
    >>> nc1
    <netCDF4.Dataset at 0x1a7dad0>
    >>> nc2 = _open_netcdf_file('file.nc')
    >>> nc2 is nc1
    True

    """
    if filename in _file_to_Dataset and mode == "r":
        # File is already open
        return _file_to_Dataset[filename]
    elif open_files_threshold_exceeded():
        # Close a random data file to make way for this one
        close_one_file()

    if mode == "a":
        if not isfile(filename):
            nc = netCDF4_Dataset(filename, "w", format=fmt)
            nc.close()
        elif filename in _file_to_Dataset:
            _close_netcdf_file(filename)
    # --- End: if

    try:
        nc = netCDF4_Dataset(filename, mode, format=fmt)
    except RuntimeError as runtime_error:
        raise RuntimeError("{0}: {1}".format(runtime_error, filename))

    if mode == "r":
        # Update the _file_to_Dataset dictionary
        _file_to_Dataset[filename] = nc

    return nc


def _close_netcdf_file(filename):
    """Close a netCDF file.

    Does nothing if the file is already closed.

    :Parameters:

        filename: `str`
            The netCDF file to be closed.

    :Returns:

        `None`

    """
    nc = _file_to_Dataset.pop(filename, None)
    if nc is not None:
        nc.close()


def _open_um_file(
    filename, aggregate=True, fmt=None, word_size=None, byte_ordering=None
):
    """Open a UM fields file or PP file and read it into a `umfile.File`
    object.

    If there is already a `umfile.File` object for the file then it is
    returned with an open file descriptor.

    :Parameters:

        filename: `str`
            The file to be opened.

    :Returns:

        `umfile.File`
            The opened file with an open file descriptor.

    """
    f = _file_to_UM.get(filename)

    if f is not None:
        if f.fd is None:
            if open_files_threshold_exceeded():
                # Close a random data array file to make way for this
                # one
                close_one_file()

            f.open_fd()
        # --- End: if
        return f

    if open_files_threshold_exceeded():
        # Close a random data array file to make way for this one
        close_one_file()

    try:
        f = File(
            path=filename,
            byte_ordering=byte_ordering,
            word_size=word_size,
            fmt=fmt,
        )
    except Exception as error:
        try:
            f.close_fd()
        except Exception:
            pass

        raise Exception(error)

    # Add a close method to the file object
    f.close = f.close_fd

    # Update the _file_to_UM dictionary
    _file_to_UM[filename] = f

    return f


def _close_um_file(filename):
    """Close a PP or UM fields file.

    Does nothing if the file is already closed.

    :Parameters:

        filename : str
            The file to be closed.

    :Returns:

        `None`

    """
    f = _file_to_UM.pop(filename, None)
    if f is not None:
        f.close_fd()



from . import Units
from .data.data import Data
from .functions import _DEPRECATION_ERROR_FUNCTION_KWARGS
from .regrid import get_cartesian_coords


def relative_vorticity(
    u, v, wrap=None, one_sided_at_boundary=False, radius=6371229.0, cyclic=None
):
    """Calculate the relative vorticity using centred finite
    differences.

    The relative vorticity of wind defined on a Cartesian domain (such
    as a plane projection) is defined as

      ζcartesian = δv/δx − δu/δy

    where x and y are points on along the 'X' and 'Y' Cartesian
    dimensions respectively; and u and v denote the 'X' and 'Y'
    components of the horizontal winds.

    If the wind field field is defined on a spherical
    latitude-longitude domain then a correction factor is included:

      ζspherical = δv/δx − δu/δy + (u/a)tan(ϕ)

    where u and v denote the longitudinal and latitudinal components
    of the horizontal wind field; a is the radius of the Earth; and ϕ
    is the latitude at each point.

    The relative vorticity is calculated using centred finite
    differences (see the *one_sided_at_boundary* parameter).

    The grid may be global or limited area. If missing values are
    present then missing values will be returned at points where the
    centred finite difference could not be calculated. The boundary
    conditions may be cyclic in longitude. The non-cyclic boundaries
    may either be filled with missing values or calculated with
    off-centre finite differences.

    Reference: H.B. Bluestein, Synoptic-Dynamic Meteorology in
    Midlatitudes, 1992, Oxford Univ. Press p113-114

    :Parameters:

        u: `Field`
            A field containing the x-wind. Must be on the same grid as
            the y-wind.

        v: `Field`
            A field containing the y-wind. Must be on the same grid as
            the x-wind.

        radius: optional
            The radius of the sphere when the winds are on a spherical
            polar coordinate domain. May be any numeric scalar object
            that can be converted to a `Data` object (which includes
            numpy array and `Data` objects). By default *radius* has a
            value of 6371229.0 metres, representing the Earth's
            radius. If units are not specified then units of metres
            are assumed.

            *Parameter example:*
              Five equivalent ways to set a radius of 6371200 metres:
              ``radius=6371200``, ``radius=numpy.array(6371200)``,
              ``radius=cf.Data(6371200)``, ``radius=cf.Data(6371200,
              'm')``, ``radius=cf.Data(6371.2, 'km')``.

        wrap: `bool`, optional
            Whether the longitude is cyclic or not. By default this is
            autodetected.

        one_sided_at_boundary: `bool`, optional
            If True then if the field is not cyclic off-centre finite
            differences are calculated at the boundaries, otherwise
            missing values are used at the boundaries.

    :Returns:

        `Field`
            The relative vorticity calculated with centred finite
            differences.

    """
    if cyclic:
        _DEPRECATION_ERROR_FUNCTION_KWARGS(
            "relative_vorticity",
            {"cyclic": cyclic},
            "Use the 'wrap' keyword instead",
        )  # pragma: no cover

    # Get the standard names of u and v
    u_std_name = u.get_property("standard_name", None)
    v_std_name = v.get_property("standard_name", None)

    # Copy u and v
    u = u.copy()
    v = v.copy()

    # Get the X and Y coordinates
    (u_x_key, u_y_key), (u_x, u_y) = get_cartesian_coords(u, "u", ("X", "Y"))
    (v_x_key, v_y_key), (v_x, v_y) = get_cartesian_coords(v, "v", ("X", "Y"))

    if not u_x.equals(v_x) or not u_y.equals(v_y):
        raise ValueError("u and v must be on the same grid.")

    # Check for lat/long
    is_latlong = (u_x.Units.islongitude and u_y.Units.islatitude) or (
        u_x.units == "degrees" and u_y.units == "degrees"
    )

    # Check for cyclicity
    if wrap is None:
        if is_latlong:
            wrap = u.iscyclic(u_x_key)
        else:
            wrap = False

    # Find the relative vorticity
    if is_latlong:
        # Save the units of the X and Y coordinates
        x_units = u_x.Units
        y_units = u_y.Units

        # Change the units of the lat/longs to radians
        radians = Units("radians")
        u_x.Units = radians
        u_y.Units = radians
        v_x.Units = radians
        v_y.Units = radians

        # Find cos and tan of latitude
        cos_lat = u_y.cos()
        tan_lat = u_y.tan()

        # Reshape for broadcasting
        u_shape = [1] * u.ndim
        u_y_index = u.get_data_axes().index(u_y_key)
        u_shape[u_y_index] = u_y.size

        v_shape = [1] * v.ndim
        v_y_index = v.get_data_axes().index(v_y_key)
        v_shape[v_y_index] = v_y.size

        # Calculate the correction term
        corr = u.copy()
        corr *= tan_lat.array.reshape(u_shape)

        # Calculate the derivatives
        v.derivative(
            v_x_key,
            wrap=wrap,
            one_sided_at_boundary=one_sided_at_boundary,
            inplace=True,
        )
        v.data /= cos_lat.array.reshape(v_shape)
        u.derivative(
            u_y_key, one_sided_at_boundary=one_sided_at_boundary, inplace=True
        )

        radius = Data.asdata(radius).squeeze()
        radius.dtype = float
        if radius.size != 1:
            raise ValueError(f"Multiple radii: radius={radius!r}")

        if not radius.Units:
            radius.override_units(Units("metres"), inplace=True)
        elif not radius.Units.equivalent(Units("metres")):
            raise ValueError(f"Invalid units for radius: {radius.Units!r}")

        # Calculate the relative vorticity. Do v-(u-corr) rather than
        # v-u+corr to be nice with coordinate reference corner cases.
        rv = v - (u - corr)
        rv.data /= radius

        # Convert the units of latitude and longitude to canonical units
        rv.dimension_coordinate("X").Units = x_units
        rv.dimension_coordinate("Y").Units = y_units

    else:
        v.derivative(
            v_x_key, one_sided_at_boundary=one_sided_at_boundary, inplace=True
        )
        u.derivative(
            u_y_key, one_sided_at_boundary=one_sided_at_boundary, inplace=True
        )

        rv = v - u

    # Convert the units of relative vorticity to canonical units
    rv.Units = Units("s-1")

    # Set the standard name if appropriate and delete the long_name
    if (u_std_name == "eastward_wind" and v_std_name == "northward_wind") or (
        u_std_name == "x_wind" and v_std_name == "y_wind"
    ):
        rv.standard_name = "atmosphere_relative_vorticity"
    else:
        rv.del_property("standard_name", None)

    rv.del_property("long_name", None)

    return rv


def histogram(*digitized):
    """Return the distribution of a set of variables in the form of an
    N-dimensional histogram.

    The number of dimensions of the histogram is equal to the number
    of field constructs provided by the *digitized* argument. Each
    such field construct defines a sequence of bins and provides
    indices to the bins that each value of one of the variables
    belongs. There is no upper limit to the number of dimensions of
    the histogram.

    The output histogram bins are defined by the exterior product of
    the one-dimensional bins of each digitized field construct. For
    example, if only one digitized field construct is provided then
    the histogram bins simply comprise its one-dimensional bins; if
    there are two digitized field constructs then the histogram bins
    comprise the two-dimensional matrix formed by all possible
    combinations of the two sets of one-dimensional bins; etc.

    An output value for an histogram bin is formed by counting the
    number cells for which the digitized field constructs, taken
    together, index that bin. Note that it may be the case that not
    all output bins are indexed by the digitized field constructs, and
    for these bins missing data is returned.

    The returned field construct will have a domain axis construct for
    each dimension of the histogram, with a corresponding dimension
    coordinate construct that defines the bin boundaries.

    .. versionadded:: 3.0.2

    .. seealso:: `cf.Field.bin`, `cf.Field.collapse`,
                 `cf.Field.digitize`, `cf.Field.percentile`,
                 `cf.Field.where`

    :Parameters:

        digitized: one or more `Field`
            One or more field constructs that contain digitized data
            with corresponding metadata, as would be output by
            `cf.Field.digitize`. Each field construct contains indices
            to the one-dimensional bins to which each value of an
            original field construct belongs; and there must be
            ``bin_count`` and ``bin_bounds`` properties as defined by
            the `cf.Field.digitize` method (and any of the extra
            properties defined by that method are also recommended).

            The bins defined by the ``bin_count`` and ``bin_bounds``
            properties are used to create a dimension coordinate
            construct for the output field construct.

            Each digitized field construct must be transformable so
            that its data is broadcastable to any other digitized
            field contruct's data. This is done by using the metadata
            constructs of the to create a mapping of physically
            compatible dimensions between the fields, and then
            manipulating the dimensions of the digitized field
            construct's data to ensure that broadcasting can occur.

    :Returns:

        `Field`
            The field construct containing the histogram.

    **Examples**

    Create a one-dimensional histogram based on 10 equally-sized bins
    that exactly span the data range:

    >>> f = cf.example_field(0)
    >>> print(f)
    Field: specific_humidity (ncvar%q)
    ----------------------------------
    Data            : specific_humidity(latitude(5), longitude(8)) 1
    Cell methods    : area: mean
    Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                    : longitude(8) = [22.5, ..., 337.5] degrees_east
                    : time(1) = [2019-01-01 00:00:00]
    >>> print(f.array)
    [[0.007 0.034 0.003 0.014 0.018 0.037 0.024 0.029]
     [0.023 0.036 0.045 0.062 0.046 0.073 0.006 0.066]
     [0.11  0.131 0.124 0.146 0.087 0.103 0.057 0.011]
     [0.029 0.059 0.039 0.07  0.058 0.072 0.009 0.017]
     [0.006 0.036 0.019 0.035 0.018 0.037 0.034 0.013]]
    >>> indices, bins = f.digitize(10, return_bins=True)
    >>> print(indices)
    Field: long_name=Bin index to which each 'specific_humidity' value belongs (ncvar%q)
    ------------------------------------------------------------------------------------
    Data            : long_name=Bin index to which each 'specific_humidity' value belongs(latitude(5), longitude(8))
    Cell methods    : area: mean
    Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                    : longitude(8) = [22.5, ..., 337.5] degrees_east
                    : time(1) = [2019-01-01 00:00:00]
    >>> print(bins.array)
    [[0.003  0.0173]
     [0.0173 0.0316]
     [0.0316 0.0459]
     [0.0459 0.0602]
     [0.0602 0.0745]
     [0.0745 0.0888]
     [0.0888 0.1031]
     [0.1031 0.1174]
     [0.1174 0.1317]
     [0.1317 0.146 ]]
    >>> h = cf.histogram(indices)
    >>> rint(h)
    Field: number_of_observations
    -----------------------------
    Data            : number_of_observations(specific_humidity(10)) 1
    Cell methods    : latitude: longitude: point
    Dimension coords: specific_humidity(10) = [0.01015, ..., 0.13885] 1
    >>> print(h.array)
    [9 7 9 4 5 1 1 1 2 1]
    >>> print(h.coordinate('specific_humidity').bounds.array)
    [[0.003  0.0173]
     [0.0173 0.0316]
     [0.0316 0.0459]
     [0.0459 0.0602]
     [0.0602 0.0745]
     [0.0745 0.0888]
     [0.0888 0.1031]
     [0.1031 0.1174]
     [0.1174 0.1317]
     [0.1317 0.146 ]]


    Create a two-dimensional histogram based on specific humidity and
    temperature bins. The temperature bins in this example are derived
    from a dummy temperature field construct with the same shape as
    the specific humidity field construct already in use:

    >>> g = f.copy()
    >>> g.standard_name = 'air_temperature'
    >>> import numpy
    >>> g[...] = numpy.random.normal(loc=290, scale=10, size=40).reshape(5, 8)
    >>> g.override_units('K', inplace=True)
    >>> print(g)
    Field: air_temperature (ncvar%q)
    --------------------------------
    Data            : air_temperature(latitude(5), longitude(8)) K
    Cell methods    : area: mean
    Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                    : longitude(8) = [22.5, ..., 337.5] degrees_east
                    : time(1) = [2019-01-01 00:00:00]

    >>> indices_t = g.digitize(5)
    >>> h = cf.histogram(indices, indices_t)
    >>> print(h)
    Field: number_of_observations
    -----------------------------
    Data            : number_of_observations(air_temperature(5), specific_humidity(10)) 1
    Cell methods    : latitude: longitude: point
    Dimension coords: air_temperature(5) = [281.1054839143287, ..., 313.9741786365939] K
                    : specific_humidity(10) = [0.01015, ..., 0.13885] 1
    >>> print(h.array)
    [[2  1  5  3  2 -- -- -- -- --]
     [1  1  2 --  1 --  1  1 -- --]
     [4  4  2  1  1  1 -- --  1  1]
     [1  1 -- --  1 -- -- --  1 --]
     [1 -- -- -- -- -- -- -- -- --]]
    >>> h.sum()
    <CF Data(): 40 1>

    """
    if not digitized:
        raise ValueError(
            "Must provide at least one 'digitized' field construct"
        )

    f = digitized[0].copy()
    f.clear_properties()

    return f.bin("sample_size", digitized=digitized)


def curl_xy(
    fx,
    fy,
    x_wrap=None,
    one_sided_at_boundary=False,
    radius=None,
):
    r"""Calculate the horizontal curl of an (X, Y) vector.

    The horizontal curl is calculated from orthogonal vector component
    fields which have dimension coordinates of X and Y, in either
    Cartesian (e.g. plane projection) or spherical polar coordinate
    systems.

    The horizontal curl of the :math:`(f_x, f_y)` vector in Cartesian
    coordinates is given by:

    .. math:: \nabla \times (f_{x}(x,y), f_{y}(x,y)) =
                \frac{\partial f_y}{\partial x}
                -
                \frac{\partial f_x}{\partial y}

    The horizontal curl of the :math:`(f_\theta, f_\phi)` vector in
    spherical polar coordinates is given by:

    .. math:: \nabla \times (f_\theta(\theta,\phi), f_\phi(\theta,\phi)) =
                \frac{1}{r \sin\theta}
                \left(
                \frac{\partial (f_\phi \sin\theta)}{\partial \theta}
                -
                \frac{\partial f_\theta}{\partial \phi}
                \right)

    where *r* is radial distance to the origin, :math:`\theta` is the
    polar angle with respect to polar axis, and :math:`\phi` is the
    azimuthal angle.

    The curl is calculated using centred finite differences apart from
    at the boundaries (see the *x_wrap* and *one_sided_at_boundary*
    parameters). If missing values are present then missing values
    will be returned at all points where a centred finite difference
    could not be calculated.

    .. versionadded:: 3.12.0

    .. seealso:: `cf.div_xy`, `cf.Field.derivative`,
                 `cf.Field.grad_xy`, `cf.Field.iscyclic`,
                 `cf.Field.laplacian_xy`

    :Parameters:

        fx, fy: `Field`
            The fields containing the X and Y vector components.

        x_wrap: `bool`, optional
            Whether the X axis is cyclic or not. By default *x_wrap*
            is set to the result of this call to the *fx* field
            construct's `~cf.Field.iscyclic`
            method:``fy.iscyclic('X')``. If the X axis is cyclic then
            centred differences at one X boundary will always use
            values from the other, regardless of the setting of
            *one_sided_at_boundary*.

            The cyclicity of the Y axis is always set to the result of
            ``fx.iscyclic('Y')``.

        one_sided_at_boundary: `bool`, optional
            If True then one-sided finite differences are calculated
            at the non-cyclic boundaries. By default missing values
            are set at non-cyclic boundaries.

        radius: optional
            Specify the radius of the latitude-longitude plane defined
            in spherical polar coordinates. The radius is that which
            would be returned by this call of the *fx* field
            construct's `cf.Field.radius` method:
            ``fx.radius(default=radius)``. The radius is defined by
            the datum of a coordinate reference construct, and if and
            only if no such radius is found then the default value
            given by the *radius* parameter is used instead. A value
            of ``'earth'`` is equivalent to a default value of 6371229
            metres.

    :Returns:

        `Field`
            The horizontal curl of the (X, Y) fields.

    **Examples**

    >>> f = cf.example_field(0)
    >>> print(f)
    Field: specific_humidity (ncvar%q)
    ----------------------------------
    Data            : specific_humidity(latitude(5), longitude(8)) 1
    Cell methods    : area: mean
    Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                    : longitude(8) = [22.5, ..., 337.5] degrees_east
                    : time(1) = [2019-01-01 00:00:00]
    >>> f[...] = 0.1
    >>> print(f.array)
    [[0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1]
     [0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1]
     [0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1]
     [0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1]
     [0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1]]
    >>> fx, fy = f.grad_xy(radius='earth', one_sided_at_boundary=True)
    >>> fx, fy
    (<CF Field: long_name=X gradient of specific_humidity(latitude(5), longitude(8)) m-1.rad-1>,
     <CF Field: long_name=Y gradient of specific_humidity(latitude(5), longitude(8)) m-1.rad-1>)
    >>> c = cf.curl_xy(fx, fy, radius='earth')
    >>> c
    <CF Field: long_name=Divergence of (long_name=X gradient of specific_humidity, long_name=Y gradient of specific_humidity)(latitude(5), longitude(8)) m-2.rad-2>
    >>> print(c.array)
    [[-- -- -- -- -- -- -- --]
     [0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0]
     [0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0]
     [0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0]
     [-- -- -- -- -- -- -- --]]
    >>> c = cf.curl_xy(fx, fy, radius='earth', one_sided_at_boundary=True)
    >>> print(c.array)
    [[0. 0. 0. 0. 0. 0. 0. 0.]
     [0. 0. 0. 0. 0. 0. 0. 0.]
     [0. 0. 0. 0. 0. 0. 0. 0.]
     [0. 0. 0. 0. 0. 0. 0. 0.]
     [0. 0. 0. 0. 0. 0. 0. 0.]]

    """
    from numpy import pi

    fx = fx.copy()
    fy = fy.copy()

    x_key, x_coord = fy.dimension_coordinate(
        "X", item=True, default=(None, None)
    )
    y_key, y_coord = fx.dimension_coordinate(
        "Y", item=True, default=(None, None)
    )

    if x_coord is None:
        raise ValueError("'fy' field has no unique 'X' dimension coordinate")

    if y_coord is None:
        raise ValueError("'fx' field has no unique 'Y' dimension coordinate")

    if x_wrap is None:
        x_wrap = fx.iscyclic(x_key)

    x_units = x_coord.Units
    y_units = y_coord.Units

    # Check for spherical polar coordinates
    latlon = (x_units.islongitude and y_units.islatitude) or (
        x_units.units == "degrees" and y_units.units == "degrees"
    )

    if latlon:
        # --------------------------------------------------------
        # Spherical polar coordinates
        # --------------------------------------------------------
        # Convert latitude and longitude units to radians, so that the
        # units of the result are nice.
        radians = Units("radians")
        x_coord.Units = radians
        y_coord.Units = radians

        # Get theta as a field that will broadcast to f, and adjust
        # its values so that theta=0 is at the north pole.
        theta = pi / 2 - fx.convert(y_key, full_domain=True)
        sin_theta = theta.sin()

        r = fx.radius(default=radius)

        term1 = (fx * sin_theta).derivative(
            y_key,
            wrap=None,
            one_sided_at_boundary=one_sided_at_boundary,
        )

        term2 = fy.derivative(
            x_key, wrap=x_wrap, one_sided_at_boundary=one_sided_at_boundary
        )

        c = (term1 - term2) / (sin_theta * r)

        # Reset latitude and longitude coordinate units
        c.dimension_coordinate("X").Units = x_units
        c.dimension_coordinate("Y").Units = y_units
    else:
        # --------------------------------------------------------
        # Cartesian coordinates
        # --------------------------------------------------------
        dfy_dx = fy.derivative(
            x_key,
            wrap=x_wrap,
            one_sided_at_boundary=one_sided_at_boundary,
        )

        dfx_dy = fx.derivative(
            y_key,
            wrap=None,
            one_sided_at_boundary=one_sided_at_boundary,
        )

        c = dfy_dx - dfx_dy

    # Set the standard name and long name
    c.set_property(
        "long_name", f"Horizontal curl of ({fx.identity()}, {fy.identity()})"
    )
    c.del_property("standard_name", None)

    return c


def div_xy(
    fx,
    fy,
    x_wrap=None,
    one_sided_at_boundary=False,
    radius=None,
):
    r"""Calculate the horizontal divergence of an (X, Y) vector.

    The horizontal divergence is calculated from orthogonal vector
    component fields which have dimension coordinates of X and Y, in
    either Cartesian (e.g. plane projection) or spherical polar
    coordinate systems.

    The horizontal divergence of the :math:`(f_x, f_y)` vector in
    Cartesian coordinates is given by:

    .. math:: \nabla \cdot (f_{x}(x,y), f_{y}(x,y)) =
                \frac{\partial f_x}{\partial x}
                +
                \frac{\partial f_y}{\partial y}

    The horizontal divergence of the :math:`(f_\theta, f_\phi)` vector
    in spherical polar coordinates is given by:

    .. math:: \nabla \cdot (f_\theta(\theta,\phi), f_\phi(\theta,\phi)) =
                \frac{1}{r \sin\theta}
                \left(
                \frac{\partial (f_\theta \sin\theta)}{\partial \theta}
                +
                \frac{\partial f_\phi}{\partial \phi}
                \right)

    where *r* is radial distance to the origin, :math:`\theta` is the
    polar angle with respect to polar axis, and :math:`\phi` is the
    azimuthal angle.

    The divergence is calculated using centred finite differences
    apart from at the boundaries (see the *x_wrap* and
    *one_sided_at_boundary* parameters). If missing values are present
    then missing values will be returned at all points where a centred
    finite difference could not be calculated.

    .. versionadded:: 3.12.0

    .. seealso:: `cf.curl_xy`, `cf.Field.derivative`,
                 `cf.Field.grad_xy`, `cf.Field.iscyclic`,
                 `cf.Field.laplacian_xy`

    :Parameters:

        fx, fy: `Field`
            The fields containing the X and Y vector components.

        x_wrap: `bool`, optional
            Whether the X axis is cyclic or not. By default *x_wrap*
            is set to the result of this call to the *fx* field
            construct's `~cf.Field.iscyclic`
            method:``fx.iscyclic('X')``. If the X axis is cyclic then
            centred differences at one X boundary will always use
            values from the other, regardless of the setting of
            *one_sided_at_boundary*.

            The cyclicity of the Y axis is always set to the result of
            ``fy.iscyclic('Y')``.

        one_sided_at_boundary: `bool`, optional
            If True then one-sided finite differences are calculated
            at the non-cyclic boundaries. By default missing values
            are set at non-cyclic boundaries.

        radius: optional
            Specify the radius of the latitude-longitude plane defined
            in spherical polar coordinates. The radius is that which
            would be returned by this call of the *fx* field
            construct's `cf.Field.radius` method:
            ``fx.radius(default=radius)``. The radius is defined by
            the datum of a coordinate reference construct, and if and
            only if no such radius is found then the default value
            given by the *radius* parameter is used instead. A value
            of ``'earth'`` is equivalent to a default value of 6371229
            metres.

    :Returns:

        `Field`
            The horizontal curl of the (X, Y) fields.

    **Examples**

    >>> f = cf.example_field(0)
    >>> print(f)
    Field: specific_humidity (ncvar%q)
    ----------------------------------
    Data            : specific_humidity(latitude(5), longitude(8)) 1
    Cell methods    : area: mean
    Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                    : longitude(8) = [22.5, ..., 337.5] degrees_east
                    : time(1) = [2019-01-01 00:00:00]
    >>> f[...] = 0.1
    >>> print(f.array)
    [[0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1]
     [0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1]
     [0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1]
     [0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1]
     [0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1]]
    >>> fx, fy = f.grad_xy(radius='earth', one_sided_at_boundary=True)
    >>> fx, fy
    (<CF Field: long_name=X gradient of specific_humidity(latitude(5), longitude(8)) m-1.rad-1>,
     <CF Field: long_name=Y gradient of specific_humidity(latitude(5), longitude(8)) m-1.rad-1>)
    >>> d = cf.div_xy(fx, fy, radius='earth')
    >>> d
    <CF Field: long_name=Divergence of (long_name=X gradient of specific_humidity, long_name=Y gradient of specific_humidity)(latitude(5), longitude(8)) m-2.rad-2>
    >>> print(d.array)
    [[-- -- -- -- -- -- -- --]
     [0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0]
     [0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0]
     [0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0]
     [-- -- -- -- -- -- -- --]]
    >>> d = cf.div_xy(fx, fy, radius='earth', one_sided_at_boundary=True)
    >>> print(d.array)
    [[0. 0. 0. 0. 0. 0. 0. 0.]
     [0. 0. 0. 0. 0. 0. 0. 0.]
     [0. 0. 0. 0. 0. 0. 0. 0.]
     [0. 0. 0. 0. 0. 0. 0. 0.]
     [0. 0. 0. 0. 0. 0. 0. 0.]]

    """
    from numpy import pi

    fx = fx.copy()
    fy = fy.copy()

    x_key, x_coord = fx.dimension_coordinate(
        "X", item=True, default=(None, None)
    )
    y_key, y_coord = fy.dimension_coordinate(
        "Y", item=True, default=(None, None)
    )

    if x_coord is None:
        raise ValueError("'fx' field has no unique 'X' dimension coordinate")

    if y_coord is None:
        raise ValueError("'fy' field has no unique 'Y' dimension coordinate")

    if x_wrap is None:
        x_wrap = fx.iscyclic(x_key)

    x_units = x_coord.Units
    y_units = y_coord.Units

    # Check for spherical polar coordinates
    latlon = (x_units.islongitude and y_units.islatitude) or (
        x_units.units == "degrees" and y_units.units == "degrees"
    )

    if latlon:
        # ------------------------------------------------------------
        # Spherical polar coordinates
        # ------------------------------------------------------------
        # Convert latitude and longitude units to radians, so that
        # the units of the result are nice.
        radians = Units("radians")
        x_coord.Units = radians
        y_coord.Units = radians

        # Get theta as a field that will broadcast to f, and adjust
        # its values so that theta=0 is at the north pole.
        theta = pi / 2 - fy.convert(y_key, full_domain=True)
        sin_theta = theta.sin()

        r = fx.radius(default=radius)

        # r_sin_theta = sin_theta * r

        term1 = (
            fx.derivative(
                x_key, wrap=x_wrap, one_sided_at_boundary=one_sided_at_boundary
            )
            # / r_sin_theta
        )

        term2 = (fy * sin_theta).derivative(
            y_key, wrap=None, one_sided_at_boundary=one_sided_at_boundary
        )  # / r_sin_theta

        d = (term1 + term2) / (sin_theta * r)  # r_sin_theta

        # Reset latitude and longitude coordinate units
        d.dimension_coordinate("X").Units = x_units
        d.dimension_coordinate("Y").Units = y_units
    else:
        # ------------------------------------------------------------
        # Cartesian coordinates
        # ------------------------------------------------------------
        term1 = fx.derivative(
            x_key, wrap=x_wrap, one_sided_at_boundary=one_sided_at_boundary
        )

        term2 = fy.derivative(
            y_key, wrap=None, one_sided_at_boundary=one_sided_at_boundary
        )

        d = term1 + term2

    # Set the standard name and long name
    d.set_property(
        "long_name",
        f"Horizontal divergence of ({fx.identity()}, {fy.identity()})",
    )
    d.del_property("standard_name", None)

    return d



from cfdm.core.functions import deepcopy

from ..data import Data
from ..functions import (
    _DEPRECATION_ERROR,
    _DEPRECATION_ERROR_DICT,
    _DEPRECATION_ERROR_KWARGS,
    _DEPRECATION_ERROR_METHOD,
)
from ..functions import atol as cf_atol
from ..functions import rtol as cf_rtol
from ..mixin_container import Container
from ..query import Query
from ..units import Units


class Properties(Container):
    """Mixin class for a container of descriptive properties.

    .. versionadded:: 3.0.0

    """

    _special_properties = ()

    def __new__(cls, *args, **kwargs):
        """Store component classes.

        .. note:: If a child class requires a different component
        classes than the ones defined here, then they must be redefined
        in the child class.

        """
        instance = super().__new__(cls)
        instance._Data = Data
        return instance

    # ----------------------------------------------------------------
    # Private attributes
    # ----------------------------------------------------------------
    @property
    def _atol(self):
        """Return the tolerance on absolute differences between real
        numbers, as returned by the `cf.atol` function.

        This is used by, for example, the `_equals` method.

        """
        return cf_atol().value

    @property
    def _rtol(self):
        """Return the tolerance on relative differences between real
        numbers, as returned by the `cf.rtol` function.

        This is used by, for example, the `_equals` method.

        """
        return cf_rtol().value

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _matching_values(self, value0, value1, units=False, basic=False):
        """Whether two values match.

        The definition of "match" depends on the types of *value0* and
        *value1*.

        :Parameters:

            value0:
                The first value to be matched.

            value1:
                The second value to be matched.

            units: `bool`, optional
                If True then the units must be the same for values to be
                considered to match. By default, units are ignored in the
                comparison.

        :Returns:

            `bool`
                Whether or not the two values match.

        """
        if value1 is None:
            return False

        if isinstance(value0, Query):
            return bool(value0.evaluate(value1))  # TODO vectors

        try:
            # value0 is a re.compile object
            return bool(value0.search(value1))
        except (AttributeError, TypeError):
            if units and isinstance(value0, str):
                return Units(value0).equals(Units(value1))

            return self._equals(value1, value0, basic=basic)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def id(self):
        """An identity for the {{class}} object.

        The `id` attribute can be used to unambiguously identify
        constructs. This can be useful when identification is not
        possible from the existing properties, either because they are
        missing or because they do not provide sufficiently unique
        information.

        In general it will only be defined if explicitly set by the
        user.

        Note that `id` is not a CF property and so is not read from,
        nor written to, datasets.

        .. seealso:: `identity`, `identities`

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.id = "foo"
        >>> f.id
        'foo'
        >>> del f.id

        """
        try:
            return self._custom["id"]
        except KeyError:
            raise AttributeError(
                f"{self.__class__.__name__} doesn't have attribute 'id'"
            )

    @id.setter
    def id(self, value):
        self._custom["id"] = value

    @id.deleter
    def id(self):
        try:
            del self._custom["id"]
        except KeyError:
            raise AttributeError(
                f"{self.__class__.__name__} doesn't have attribute 'id'"
            )

    # ----------------------------------------------------------------
    # CF properties
    # ----------------------------------------------------------------
    @property
    def calendar(self):
        """The calendar CF property.

        The calendar used for encoding time data. See
        http://cfconventions.org/latest.html for details.

        **Examples**

        >>> f.calendar = 'noleap'
        >>> f.calendar
        'noleap'
        >>> del f.calendar

        >>> f.set_property('calendar', 'proleptic_gregorian')
        >>> f.has_property('calendar')
        True
        >>> f.get_property('calendar')
        'proleptic_gregorian'
        >>> f.del_property('calendar')

        """
        return self.get_property("calendar", default=AttributeError())

    @calendar.setter
    def calendar(self, value):
        self.set_property("calendar", value, copy=False)

    @calendar.deleter
    def calendar(self):
        self.del_property("calendar", default=AttributeError())

    @property
    def comment(self):
        """The comment CF property.

        Miscellaneous information about the data or methods used to
        produce it. See http://cfconventions.org/latest.html for details.

        **Examples**

        >>> f.comment = 'This simulation was done on an HP-35 calculator'
        >>> f.comment
        'This simulation was done on an HP-35 calculator'
        >>> del f.comment

        >>> f.set_property('comment', 'a comment')
        >>> f.has_property('comment')
        True
        >>> f.get_property('comment')
        'a comment'
        >>> f.del_property('comment')

        """
        return self.get_property("comment", default=AttributeError())

    @comment.setter
    def comment(self, value):
        self.set_property("comment", value, copy=False)

    @comment.deleter
    def comment(self):
        self.del_property("comment", default=AttributeError())

    @property
    def history(self):
        """The history CF property.

        A list of the applications that have modified the original
        data. See http://cfconventions.org/latest.html for details.

        **Examples**

        >>> f.history = 'created on 2012/10/01'
        >>> f.history
        'created on 2012/10/01'
        >>> del f.history

        >>> f.set_property('history', 'created on 2012/10/01')
        >>> f.has_property('history')
        True
        >>> f.get_property('history')
        'created on 2012/10/01'
        >>> f.del_property('history')

        """
        return self.get_property("history", default=AttributeError())

    @history.setter
    def history(self, value):
        self.set_property("history", value, copy=False)

    @history.deleter
    def history(self):
        self.del_property("history", default=AttributeError())

    @property
    def leap_month(self):
        """The leap_month CF property.

        Specifies which month is lengthened by a day in leap years for a
        user defined calendar. See http://cfconventions.org/latest.html
        for details.

        **Examples**

        >>> f.leap_month = 2
        >>> f.leap_month
        2
        >>> del f.leap_month

        >>> f.set_property('leap_month', 11)
        >>> f.has_property('leap_month')
        True
        >>> f.get_property('leap_month')
        11
        >>> f.del_property('leap_month')

        """
        return self.get_property("leap_month", default=AttributeError())

    @leap_month.setter
    def leap_month(self, value):
        self.set_property("leap_month", value, copy=False)

    @leap_month.deleter
    def leap_month(self):
        self.del_property("leap_month", default=AttributeError())

    @property
    def leap_year(self):
        """The leap_year CF property.

        Provides an example of a leap year for a user defined calendar. It
        is assumed that all years that differ from this year by a multiple
        of four are also leap years. See
        http://cfconventions.org/latest.html for details.

        **Examples**

        >>> f.leap_year = 1984
        >>> f.leap_year
        1984
        >>> del f.leap_year

        >>> f.set_property('leap_year', 1984)
        >>> f.has_property('leap_year')
        True
        >>> f.get_property('leap_year')
        1984
        >>> f.del_property('leap_year')

        """
        return self.get_property("leap_year", default=AttributeError())

    @leap_year.setter
    def leap_year(self, value):
        self.set_property("leap_year", value)

    @leap_year.deleter
    def leap_year(self):
        self.del_property("leap_year", default=AttributeError())

    @property
    def long_name(self):
        """The long_name CF property.

        A descriptive name that indicates a nature of the data. This name
        is not standardized. See http://cfconventions.org/latest.html for
        details.

        **Examples**

        >>> f.long_name = 'zonal_wind'
        >>> f.long_name
        'zonal_wind'
        >>> del f.long_name

        >>> f.set_property('long_name', 'surface air temperature')
        >>> f.has_property('long_name')
        True
        >>> f.get_property('long_name')
        'surface air temperature'
        >>> f.del_property('long_name')

        """
        return self.get_property("long_name", default=AttributeError())

    @long_name.setter
    def long_name(self, value):
        self.set_property("long_name", value, copy=False)

    @long_name.deleter
    def long_name(self):
        self.del_property("long_name", default=AttributeError())

    @property
    def month_lengths(self):
        """The month_lengths CF property.

        Specifies the length of each month in a non-leap year for a user
        defined calendar. See http://cfconventions.org/latest.html for
        details.

        Stored as a tuple but may be set as any array-like object.

        **Examples**

        >>> f.month_lengths = numpy.array(
        ...     [34, 31, 32, 30, 29, 27, 28, 28, 28, 32, 32, 34])
        >>> f.month_lengths
        (34, 31, 32, 30, 29, 27, 28, 28, 28, 32, 32, 34)
        >>> del f.month_lengths

        >>> f.set_property('month_lengths',
        ...                [34, 31, 32, 30, 29, 27, 28, 28, 28, 32, 32, 34])
        >>> f.has_property('month_lengths')
        True
        >>> f.get_property('month_lengths')
        (34, 31, 32, 30, 29, 27, 28, 28, 28, 32, 32, 34)
        >>> f.del_property('month_lengths')

        """
        return self.get_property("month_lengths", default=AttributeError())

    @month_lengths.setter
    def month_lengths(self, value):
        self.set_property("month_lengths", tuple(value), copy=False)

    @month_lengths.deleter
    def month_lengths(self):
        self.del_property("month_lengths", default=AttributeError())

    @property
    def standard_name(self):
        """The standard_name CF property.

        A standard name that references a description of a data in the
        standard name table
        (http://cfconventions.org/standard-names.html). See
        http://cfconventions.org/latest.html for details.

        **Examples**

        >>> f.standard_name = 'time'
        >>> f.standard_name
        'time'
        >>> del f.standard_name

        >>> f.set_property('standard_name', 'time')
        >>> f.has_property('standard_name')
        True
        >>> f.get_property('standard_name')
        'time'
        >>> f.del_property('standard_name')

        """
        return self.get_property("standard_name", default=AttributeError())

    @standard_name.setter
    def standard_name(self, value):
        self.set_property("standard_name", value, copy=False)

    @standard_name.deleter
    def standard_name(self):
        self.del_property("standard_name", default=AttributeError())

    @property
    def units(self):
        """The units CF property.

        The units of the data. The value of the `units` property is a
        string that can be recognized by UNIDATA's Udunits package
        (http://www.unidata.ucar.edu/software/udunits). See
        http://cfconventions.org/latest.html for details.

        **Examples**

        >>> f.units = 'K'
        >>> f.units
        'K'
        >>> del f.units

        >>> f.set_property('units', 'm.s-1')
        >>> f.has_property('units')
        True
        >>> f.get_property('units')
        'm.s-1'
        >>> f.del_property('units')

        """
        return self.get_property("units", default=AttributeError())

    @units.setter
    def units(self, value):
        self.set_property("units", value, copy=False)

    @units.deleter
    def units(self):
        self.del_property("units", default=AttributeError())

    @property
    def valid_max(self):
        """The valid_max CF property.

        The largest valid value of the data. See
        http://cfconventions.org/latest.html for details.

        **Examples**

        >>> f.valid_max = 100.0
        >>> f.valid_max
        100.0
        >>> del f.valid_max

        >>> f.set_property('valid_max', 100.0)
        >>> f.has_property('valid_max')
        True
        >>> f.get_property('valid_max')
        100.0
        >>> f.del_property('valid_max')

        """
        return self.get_property("valid_max", default=AttributeError())

    @valid_max.setter
    def valid_max(self, value):
        self.set_property("valid_max", value)

    @valid_max.deleter
    def valid_max(self):
        self.del_property("valid_max", default=AttributeError())

    @property
    def valid_min(self):
        """The valid_min CF property.

        The smallest valid value of the data. See
        http://cfconventions.org/latest.html for details.

        **Examples**

        >>> f.valid_min = 8.0
        >>> f.valid_min
        8.0
        >>> del f.valid_min

        >>> f.set_property('valid_min', 8.0)
        >>> f.has_property('valid_min')
        True
        >>> f.get_property('valid_min')
        8.0
        >>> f.del_property('valid_min')

        """
        return self.get_property("valid_min", default=AttributeError())

    @valid_min.setter
    def valid_min(self, value):
        self.set_property("valid_min", value)

    @valid_min.deleter
    def valid_min(self):
        self.del_property("valid_min", default=AttributeError())

    @property
    def valid_range(self):
        """The valid_range CF property.

        The smallest and largest valid values the data. See
        http://cfconventions.org/latest.html for details.

        Stored as a tuple but may be set as any array-like object.

        **Examples**

        >>> f.valid_range = numpy.array([100., 400.])
        >>> f.valid_range
        (100.0, 400.0)
        >>> del f.valid_range

        >>> f.set_property('valid_range', [100.0, 400.0])
        >>> f.has_property('valid_range')
        True
        >>> f.get_property('valid_range')
        (100.0, 400.0)
        >>> f.del_property('valid_range')

        """
        return tuple(
            self.get_property("valid_range"), default=AttributeError()
        )

    @valid_range.setter
    def valid_range(self, value):
        self.set_property("valid_range", tuple(value), copy=False)

    @valid_range.deleter
    def valid_range(self):
        self.del_property("valid_range", default=AttributeError())

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def get_property(self, prop, default=ValueError()):
        """Get a CF property.

        .. versionadded:: 3.0.0

        .. seealso:: `clear_properties`, `del_property`, `has_property`,
                     `properties`, `set_property`

        :Parameters:

            prop: `str`
                The name of the CF property.

                *Parameter example:*
                  ``prop='long_name'``

            default: optional
                Return the value of the *default* parameter if the
                property has not been set.

                {{default Exception}}

        :Returns:

                The value of the named property or the default value, if
                set.

        **Examples**

        >>> f = cf.{{class}}()
        >>> f.set_property('project', 'CMIP7')
        >>> f.has_property('project')
        True
        >>> f.get_property('project')
        'CMIP7'
        >>> f.del_property('project')
        'CMIP7'
        >>> f.has_property('project')
        False
        >>> print(f.del_property('project', None))
        None
        >>> print(f.get_property('project', None))
        None

        """
        # Get a special property
        if prop in self._special_properties:
            try:
                return getattr(self, prop)
            except AttributeError as error:
                if default is None:
                    return

                return self._default(default, error)

        # Still here? Then get a non-special property
        return super().get_property(prop, default=default)

    def has_property(self, prop):
        """Whether a property has been set.

        .. versionadded:: 3.0.0

        .. seealso:: `clear_properties`, `del_property`, `get_property`,
                     `properties`, `set_property`

        :Parameters:

            prop: `str`
                The name of the property.

                *Parameter example:*
                   ``prop='long_name'``

        :Returns:

             `bool`
                `True` if the property has been set, otherwise `False`.

        **Examples**

        >>> f = cf.{{class}}()
        >>> f.set_property('project', 'CMIP7')
        >>> f.has_property('project')
        True
        >>> f.get_property('project')
        'CMIP7'
        >>> f.del_property('project')
        'CMIP7'
        >>> f.has_property('project')
        False
        >>> print(f.del_property('project', None))
        None
        >>> print(f.get_property('project', None))
        None

        """
        # Get a special property
        if prop in self._special_properties:
            return hasattr(self, prop)

        # Still here? Then get a non-special property
        return super().has_property(prop)

    def del_property(self, prop, default=ValueError()):
        """Remove a property.

        .. versionadded:: 3.0.0

        .. seealso:: `clear_properties`, `get_property`, `has_property`,
                     `properties`, `set_property`

        :Parameters:

            prop: `str`
                The name of the property.

                *Parameter example:*
                  ``prop='long_name'``

            default: optional
                Return the value of the *default* parameter if the
                property has not been set.

                {{default Exception}}

         :Returns:

                The removed property.

        **Examples**

        >>> f = cf.{{class}}()
        >>> f.set_property('project', 'CMIP7')
        >>> f.has_property('project')
        True
        >>> f.get_property('project')
        'CMIP7'
        >>> f.del_property('project')
        'CMIP7'
        >>> f.has_property('project')
        False
        >>> print(f.del_property('project', None))
        None
        >>> print(f.get_property('project', None))
        None

        """
        # Get a special attribute
        if prop in self._special_properties:
            try:
                out = getattr(self, prop)
            except AttributeError as error:
                if default is None:
                    return

                return self._default(default, error)
            else:
                delattr(self, prop)
                return out

        # Still here? Then del a non-special attribute
        return super().del_property(prop, default=default)

    def match_by_identity(self, *identities):
        """Whether or not the construct identity satisfies conditions.

        .. versionadded:: 3.0.0

        .. seealso:: `match`, `match_by_property`, `match_by_ncvar`

        :Parameters:

            identities: optional
                Define conditions on the construct identities by one or
                more of

                * A construct identity. TODO

                  {{construct selection identity}}

                *Parameter example:*
                  ``'latitude'``

                *Parameter example:*
                  ``'T'

                *Parameter example:*
                  ``'long_name=Cell Area'``

                *Parameter example:*
                  ``'cellmeasure1'``

                *Parameter example:*
                  ``'measure:area'``

                *Parameter example:*
                  ``cf.eq('time')'``

                *Parameter example:*
                  ``re.compile('^lat')``

                *Parameter example:*
                  To match identities of "T", or any that start with
                  "lat": ``'T', re.compile('^lat')``

        :Returns:

            `bool`
                Whether or not at least one of the conditions are met.

        **Examples**

        >>> f.match_by_identity('time')

        >>> f.match_by_identity(re.compile('^air'))

        >>> f.match_by_identity('air_pressure', 'air_temperature')

        >>> f.match_by_identity('ncvar%t')

        """
        if not identities:
            return True

        ok = False
        for value1 in self.identities(generator=True):
            for value0 in identities:
                ok = self._matching_values(value0, value1, basic=True)
                if ok:
                    break

            if ok:
                break

        return ok

    def match_by_ncvar(self, *ncvars):
        """Whether or not the netCDF variable name satisfies conditions.

        .. versionadded:: 3.0.0

        .. seealso:: `match`, `match_by_identity`, `match_by_property`

        :Parameters:

            ncvars: optional
                Define one or more conditions on the netCDF variable name.

                A netCDF variable name is specified by a string
                (e.g. ``'lat'``), etc.); a `Query` object
                (e.g. ``cf.eq('lon')``); or a compiled regular expression
                (e.g. ``re.compile('^t')``) that is compared with the
                construct's netCDF variable name via `re.search`.

                The condition is satisfied if the netCDF variable name, as
                returned by the `nc_get_variable` method, exists and
                equals the condition value.

        :Returns:

            `bool`
                Whether or not at least one of the conditions are met.

        **Examples**

        >>> f.nc_get_variable()
        'time'
        >>> f.match_by_ncvar('time')
        True
        >>> f.match_by_ncvar('t', 'time')
        True
        >>> f.match_by_ncvar('t')
        False
        >>> f.match_by_ncvar()
        True
        >>> import re
        >>> f.match_by_ncvar(re.compile('^t'))
        True

        """
        if not ncvars:
            return True

        try:
            ncvar = self.nc_get_variable(None)
        except AttributeError:
            ncvar = None

        if ncvar is None:
            return False

        ok = False
        for value0 in ncvars:
            ok = self._matching_values(value0, ncvar, basic=True)
            if ok:
                break

        return ok

    def match_by_property(self, *mode, **properties):
        """Whether or not properties satisfy conditions.

        .. versionadded:: 3.0.0

        .. seealso:: `match`, `match_by_identity`, `match_by_ncvar`

        :Parameters:

            mode: optional
                Define the behaviour when multiple conditions are
                provided.

                By default (or if the *mode* parameter is ``'and'``) the
                match is `True` if the field construct satisfies all of
                the given conditions, but if the *mode* parameter is
                ``'or'`` then the match is `True` when at least one of the
                conditions is satisfied.

            properties: optional
                Define conditions on properties.

                Properties are identified by the name of a keyword
                parameter.

                The keyword parameter value defines a condition to be
                applied to the property, and is either a string
                (e.g. ``'latitude'``); a `Query` object
                (e.g. ``cf.eq('longitude')``); or a compiled regular
                expression (e.g. ``re.compile('^atmosphere')``) that is
                compared with the property value via `re.search`.

                If the value is `None` then the condition is one of
                existence and is satisfied if the property exists,
                regardless of its value. Otherwise the condition is
                satisfied if the property equals the value.

                *Parameter example:*
                  To see if there is a 'standard_name' property of 'time':
                  ``standard_name='time'``.

                *Parameter example:*
                  To see if there is a 'valid_min' property of that is
                  less than -999: ``valid_min=cf.lt('-999)``.

                *Parameter example:*
                  To see if there is a 'long_name' property with any
                  value: ``long_name=None``.

        :Returns:

            `bool`
                Whether or not the conditions are met.

        **Examples**

        >>> f.match_by_property(standard_name='longitude')

        >>> f.match_by_property(
        ...     standard_name='longitude', foo=cf.set(['bar', 'not_bar']))

        >>> f.match_by_property(long_name=re.compile('^lon'))

        """
        _or = False

        # Parse mode
        if mode:
            if len(mode) > 1:
                raise ValueError("Can provide at most one positional argument")

            x = mode[0]
            if x == "or":
                _or = True
            elif x != "and":
                raise ValueError(
                    "Positional argument, if provided, must one of 'or', "
                    "'and'"
                )

        if not properties:
            return True

        self_properties = self.properties()

        ok = True
        for name, value0 in properties.items():
            if value0 is None:
                ok = name in self_properties
            else:
                value1 = self_properties.get(name)
                ok = self._matching_values(
                    value0, value1, units=(name == "units")
                )

            if _or:
                if ok:
                    break
            elif not ok:
                break

        return ok

    def properties(self):
        """Return all properties.

        .. seealso:: `clear_properties`, `get_property`, `has_property`
                     `set_properties`

        :Returns:

            `dict`
                The properties.

        **Examples**

        >>> f.properties()
        {}
        >>> f.set_properties(
        ...     {'standard_name': 'air_pressure', 'long_name': 'Air Pressure'})
        >>> f.properties()
        {'standard_name': 'air_pressure',
         'foo': 'bar',
         'long_name': 'Air Pressure'}
        >>> f.set_properties({'standard_name': 'air_pressure', 'foo': 'bar'})
        >>> f.properties()
        {'standard_name': 'air_pressure',
         'foo': 'bar',
         'long_name': 'Air Pressure'}
        >>> f.clear_properties()
        {'standard_name': 'air_pressure',
         'foo': 'bar',
         'long_name': 'Air Pressure'}
        >>> f.properties()
        {}

        """
        out = super().properties()

        for prop in self._special_properties:
            value = getattr(self, prop, None)
            if value is None:
                out.pop(prop, None)
            else:
                out[prop] = value

        return out

    def set_properties(self, properties, copy=True):
        """Set properties.

        .. versionadded:: 3.0.0

        .. seealso:: `clear_properties`, `properties`, `set_property`

        :Parameters:

            properties: `dict`
                Store the properties from the dictionary supplied.

                *Parameter example:*
                  ``properties={'standard_name': 'altitude', 'foo': 'bar'}``

            copy: `bool`, optional
                If `False` then any property values provided by the
                *properties* parameter are not copied before insertion. By
                default they are deep copied.

        :Returns:

            `None`

        **Examples**

        >>> f.properties()
        {}
        >>> f.set_properties(
        ...     {'standard_name': 'air_pressure', 'long_name': 'Air Pressure'})
        >>> f.properties()
        {'standard_name': 'air_pressure',
         'foo': 'bar',
         'long_name': 'Air Pressure'}
        >>> f.set_properties({'standard_name': 'air_pressure', 'foo': 'bar'})
        >>> f.properties()
        {'standard_name': 'air_pressure',
         'foo': 'bar',
         'long_name': 'Air Pressure'}
        >>> f.clear_properties()
        {'standard_name': 'air_pressure',
         'foo': 'bar',
         'long_name': 'Air Pressure'}
        >>> f.properties()
        {}

        """
        super().set_properties(properties, copy=copy)

        for prop in self._special_properties:
            value = properties.get(prop)
            if value is not None:
                setattr(self, prop, value)

    def set_property(self, prop, value, copy=True):
        """Set a property.

        .. versionadded:: 3.0.0

        .. seealso:: `clear_properties`, `del_property`, `get_property`,
                     `has_property`, `properties`

        :Parameters:

            prop: `str`
                The name of the property to be set.

            value:
                The value for the property.

            copy: `bool`, optional
                If `True` then set a deep copy of *value*.

        :Returns:

             `None`

        **Examples**

        >>> f = cf.{{class}}()
        >>> f.set_property('project', 'CMIP7')
        >>> f.has_property('project')
        True
        >>> f.get_property('project')
        'CMIP7'
        >>> f.del_property('project')
        'CMIP7'
        >>> f.has_property('project')
        False
        >>> print(f.del_property('project', None))
        None
        >>> print(f.get_property('project', None))
        None

        """
        # Get a special property
        if prop in self._special_properties:
            if copy:
                value = deepcopy(value)

            setattr(self, prop, value)
            return

        # Still here? Then set a non-special property
        return super().set_property(prop, value, copy=copy)

    # ----------------------------------------------------------------
    # Aliases
    # ----------------------------------------------------------------
    def match(self, *identities, **kwargs):
        """Alias for `match_by_identity`."""
        if kwargs:
            _DEPRECATION_ERROR_KWARGS(
                self, "match", kwargs, "Use 'match_by_*' methods instead."
            )  # pragma: no cover

        if identities and isinstance(identities[0], (list, tuple, set)):
            _DEPRECATION_ERROR(
                f"Use of a {identities[0].__class__.__name__!r} for "
                "identities has been deprecated. Use the "
                "* operator to unpack the arguments instead."
            )  # pragma: no cover

        for i in identities:
            if isinstance(i, dict):
                _DEPRECATION_ERROR_DICT(
                    "Use 'match_by_*' methods instead."
                )  # pragma: no cover

        return self.match_by_identity(*identities)

    # ----------------------------------------------------------------
    # Deprecated attributes and methods
    # ----------------------------------------------------------------
    def setprop(self, *args, **kwargs):
        """Deprecated at version 3.0.0, use method `set_property`
        instead."""
        _DEPRECATION_ERROR_METHOD(
            self, "setprop", "Use method 'set_property' instead"
        )  # pragma: no cover

    def delprop(self, prop):
        """Deprecated at version 3.0.0, use method `del_property`
        instead."""
        _DEPRECATION_ERROR_METHOD(
            self, "delprop", "Use method 'del_property' instead"
        )  # pragma: no cover

    def hasprop(self, prop):
        """Deprecated at version 3.0.0, use method `has_property`
        instead."""
        _DEPRECATION_ERROR_METHOD(
            self, "hasprop", "Use method 'has_property' instead"
        )  # pragma: no cover

    def getprop(self, prop):
        """Deprecated at version 3.0.0, use method `get_property`
        instead."""
        _DEPRECATION_ERROR_METHOD(
            self, "getprop", "Use method 'get_property' instead"
        )  # pragma: no cover
